from enum import Enum
from dataclasses import dataclass

class SocketEvent(Enum):
    CLIENT_CONNECT      = 1
    CLIENT_RECEIVE      = 2

    COMMAND_RECV        = 3
    AUDIO_RECV           = 4

class SocketCommand(Enum):
    DISCONNECTED        = b''
    ACK                 = int(1).to_bytes(1, 'big', signed=False)
    LISTEN              = int(2).to_bytes(1, 'big', signed=False)
    RECORD              = int(4).to_bytes(1, 'big', signed=False)
    PLAY                = int(5).to_bytes(1, 'big', signed=False)
    TEXT_SO_FAR         = int(6).to_bytes(1, 'big', signed=False)
    AUDIO               = int(7).to_bytes(1, 'big', signed=False)
    RESULT              = int(8).to_bytes(1, 'big', signed=False)
    STOP                = int(254).to_bytes(1, 'big', signed=False)

@dataclass
class SocketConfig:
    host: str
    port: int = 50007
    ping: int = 5

    @classmethod
    def FromJSON(cls, json):
        return cls(
            json["host"],
            json["port"] if "port" in json else cls.port,
            json["ping"] if "ping" in json else cls.ping,
        )

def get_or_default(json: dict, name: str, default):
    try:
        return json[name]
    except:
        return default


from enum import Enum
from dataclasses import dataclass

class BLESelector(Enum):
    NONE = 0
    SENSORS = 1
    AUDIO = 2

class BLECommands(Enum):
    WRITE = 1
    READ = 2

class BLECommandWord(Enum):
    UPDATE_TOUCH1       = 0x41
    UPDATE_TOUCH2       = 0x42
    WRITE_TOUCH         = 0x43
    START               = 0x00
    STOP                = 0x01
    STATUS              = 0xFF

class BLECommandMask(Enum):
    SENSORFUSION        = b'\x00\x00\x01\x00'
    ECOMPASS            = b'\x00\x00\x00\x40'

@dataclass
class BLECommand:
    command: BLECommands
    characteristic: str
    payload: bytes | None = None


from os import path
from dataclasses import dataclass
from enum import Enum

SAMPLE_RATE = 16000
SAMPLE_WIDTH = 2
N_CHANNELS = 1
AUDIO_FRAME_LENGTH = 80

@dataclass
class TactigonSpeechResponse:
    complete: bool
    payload: dict

@dataclass
class HotWord:
    word: str
    boost: int = 1

    @classmethod
    def FromJSON(cls, config):
        return cls(config["word"], config["boost"])
    
    @classmethod
    def FromBytes(cls, data: bytes):
        return cls(
            "".join([chr(c) for c in data[1::]]),
            int.from_bytes(data[0:1])
        )
    
    def toJSON(self) -> object:
        return {
            "word": self.word,
            "boost": self.boost
        }
    
    def toBytes(self) -> bytes:
        return self.boost.to_bytes(1, 'big') + bytes([ord(c) for c in self.word])

class TSpeech:
    hotwords: list[HotWord]
    children: "TSpeechObject | None"

    def __init__(self, hotwords: list[HotWord] | HotWord, children: "TSpeechObject | None" = None):
        self.hotwords = hotwords if isinstance(hotwords, list) else [hotwords]
        self.children = children

    @classmethod
    def FromJSON(cls, json_obj, feedback_audio_path = ""):
        try:
            children = json_obj["children"]
        except:
            children = None

        return cls(
            [HotWord.FromJSON(hw) for hw in json_obj["hotwords"]],
            children=TSpeechObject.FromJSON(children, feedback_audio_path) if children else None,
        )
    
    @classmethod
    def FromBytes(cls, data: bytes, level: int = 0):
        end = child_index = data.find(b"|")

        if end == -1:
            end = len(data)

        print(child_index, data, data[:end], data[child_index+1:])
        return cls(
            [HotWord.FromBytes(hw) for hw in data[:end].split(b',')],
            TSpeechObject.FromBytes(data[child_index+1:], level + 1) if child_index > -1 else None
        )

    @property
    def has_children(self):
        return self.children
    
    def toJSON(self) -> dict:
        return {
            "hotwords": [hw.toJSON() for hw in self.hotwords],
            "children": self.children.toJSON() if self.children else None
        }
    
    def toBytes(self, level: int = 0):
        return b','.join([hw.toBytes() for hw in self.hotwords]) + (b'|' + self.children.toBytes(level + 1) if self.children else b'')

class TSpeechObject:
    t_speech: list[TSpeech]
    feedback: str

    def __init__(self, t_speech: list[TSpeech], feedback: str = ""):
        self.t_speech = t_speech
        self.feedback = feedback

    @classmethod
    def FromJSON(cls, json_obj, feedback_audio_path = ""):

        try:
            feedback = path.join(feedback_audio_path, json_obj["feedback"])
        except:
            feedback = ""

        return cls(
            [TSpeech.FromJSON(t, feedback_audio_path) for t in json_obj["t_speech"]],
            feedback=feedback
        )
    
    @classmethod
    def FromBytes(cls, data: bytes, level: int = 0):
        sep = b'-' + level.to_bytes(1, 'big')
        return cls(
            [TSpeech.FromBytes(ts, level) for ts in data.rsplit(sep)]
        )
    
    def toJSON(self) -> dict:
        return {
            "t_speech": [ts.toJSON() for ts in self.t_speech],
            "feedback": self.feedback
        }
    
    def toBytes(self, level: int = 0) -> bytes:
        return (b'-'+ level.to_bytes(1, 'big')).join([ts.toBytes(level) for ts in self.t_speech])

class TStreamStatus(Enum):
    STREAMING = 1
    STOPPED = 2

@dataclass
class Transcription:
    text: str
    path: list[HotWord]
    time: float
    timeout: bool

    @classmethod
    def FromJSON(cls, json: dict):
        return cls(
            text=json.get("text", ""),
            path=[HotWord.FromJSON(hw) for hw in json.get("path", [])],
            time=json.get("time", 0),
            timeout=json.get("payload", True)
        )

    def toJSON(self) -> dict:
        return {
            "text": self.text,
            "path": [hw.toJSON() for hw in self.path] if self.path else [],
            "time": self.time,
            "timeout": self.timeout
        }
    
@dataclass
class VoiceConfig:
    model: str
    scorer: str | None = None
    beam_width: int = 1024
    voice_timeout: int = 8
    silence_timeout: int = 3

    stop_hotword: HotWord | None = None

    @property
    def model_full_path(self) -> str:
        return path.join(self.model)
    
    @property
    def scorer_full_path(self) -> str | None:
        if self.scorer is None:
            return None
        
        return path.join(self.scorer)

    @classmethod
    def FromJSON(cls, json: dict):
        return cls(
            json["model"],
            json.get("scrore", None),
            json.get("beam_width", 1024),
            json.get("voice_timeout", 8),
            json.get("silence_timeout", 3),
        )
    
    def toJSON(self) -> dict:
        return {
            "model": self.model,
            "scorer": self.scorer,
            "beam_width": self.beam_width,
            "voice_timeout": self.voice_timeout,
            "silence_timeout": self.silence_timeout
        }


import threading
import io
from typing import Optional, Any


from .tskin import TSkin
from .connection.listener import Listener
from .models.tskin import TSkinConfig
from .models.audio import TSpeechObject, Transcription, HotWord
from .notifications.audio import TRIGGER_NOTIFICATION_AUDIO
from .models.socket import SocketConfig, SocketCommand

class TSkinSocket(TSkin):
    AUDIO_PACKET_LEN: int = 50
    config: TSkinConfig
    server_config: SocketConfig

    _listener: Listener
    _command: SocketCommand | None
    _audio_data: list[bytes]

    def __init__(self, config: TSkinConfig, socket_config: SocketConfig, debug: bool = False):
        TSkin.__init__(self, config, debug)

        self.server_config = socket_config
        self._listener = Listener(self.server_config.host, self.server_config.port)
        self._listener.task = self.on_response

        self._command = None
        self._audio_data = []

    def handle_audio(self, char, data: bytearray):
        self._audio_data.append(self.adpcm_engine.extract_data(data))

        if len(self._audio_data) >= self.AUDIO_PACKET_LEN:
            self._listener.send(SocketCommand.AUDIO, self._audio_data)
            self._audio_data.clear()

    def on_response(self, command: SocketCommand, payload: dict):
        if command == SocketCommand.ACK:
            self._logger.info("Ready to stream audio!")

        if command == SocketCommand.RESULT:
            if "transcription" in payload:
                self._transcription = Transcription.FromJSON(payload["transcription"])
                self._logger.debug("Transcription %s", self._transcription)

        if command in [SocketCommand.RESULT, SocketCommand.DISCONNECTED, SocketCommand.STOP]:
            self.select_sensors()
            self._command = None     

    def send_command(self, command: SocketCommand, payload: dict = {}):
        self._listener.send(command, payload)

    def start(self):
        TSkin.start(self)
        self._listener.start()

    def join(self, timeout: Optional[float] = None):
        self.stop()
        self._listener.join(timeout)
        TSkin.join(self, timeout)

    @property
    def is_listening(self) -> bool:
        return self._command == SocketCommand.LISTEN

    def listen(self, speech: Optional[TSpeechObject] = None) -> bool:
        if self._command is not None:
            return False
        
        self.send_command(SocketCommand.LISTEN, speech.toJSON() if speech else {})
        self.select_audio()
        self._command = SocketCommand.LISTEN

        return True

    def play(self, filename: str) -> bool:
        if self._command is not None:
            return False
        
        self.send_command(SocketCommand.PLAY, {"filename":filename})
        
        return True
       
    @property
    def is_recording(self) -> bool:
        return self._command == SocketCommand.RECORD

    def record(self, filename: str, seconds: float = 5) -> bool:
        if self._command is not None:
            return False
        
        self.send_command(SocketCommand.RECORD, {"filename":filename, "seconds":seconds})
        self.select_audio()
        self._command = SocketCommand.RECORD

        return True

    def stop(self):
        self.send_command(SocketCommand.STOP)



import base64

TRIGGER_NOTIFICATION_AUDIO = base64.b64decode(
    "UklGRqRBAABXQVZFZm10IBAAAAABAAEARKwAAIhYAQACABAAZGF0YVZBAACFBBQFjAX/BVsGggae"
    "Bq8GzwbSBqYGbgZUBiQGugVMBfcErQRCBKgDGwO2AjkCzAFfAcoAMwCN/+n+Nv6O/e78W/zk+3X7"
    "Nfsb+//64Pro+h/7QftE+0H7L/vk+nP6w/n++Dr4dvee9t/1S/X89PP09/Qm9Xz1IfbQ9ov3Rvg0"
    "+Q36t/pV+/H7dPzV/Br9dP3o/UH+fP6//hr/pP9ZABoB9wEEA/4D/QTbBbYGfAcgCGQIgAiKCG4I"
    "DAi/B1sH4AZ2BiwGvQWABXIFZwVCBSUF8gTEBKUEbgRABCQEywM0A3kCmwGcAKn/t/7i/XT9Nf3y"
    "/Nn86PwV/VT9cP1s/YL9mf2H/Y/9d/1F/SX98vyj/H78dfw0/An86/vN+8v74fvU+/f7TvyX/PT8"
    "X/2k/d79I/4i/iH+Sf5U/kn+Vv5y/p/+Ff9b/5D/7/8hADEAVgBtAIkAxwDnAPQANAFzAYYBswHL"
    "Ab0B0wHiAc0ByAG9AWAB7AB8APz/kf8p/4r+Ff7W/bP9yf0y/pT+E/+Q/97/KACCAJ0AowCSAFEA"
    "///P/3H//P62/mz+Ff7s/e/9IP6Q/hj/of9FAOEAbAEKAoMCxQL5AvYCowJGAsoBLQGhADEAl/8W"
    "/8D+YP47/kD+Sv5l/rD+Af9W/6n/9/9TANYAOQGoAQ8CawKyAt0ClgIGAloBdgCW/wL/ov4+/hf+"
    "F/4t/kj+ff7C/gP/Hf8w/07/Sv8k/+H+dv7P/SD9ZPyS+8L6Dvp8+Q75y/jI+Dn5+/nO+qv7vvzg"
    "/Rz/fQDgAfMCsgNZBOMEEgXqBKAEQQTAAzIDfQK3AUQBfQEjAj0DIQXQB98K/Q01ER0UaBYZGFgZ"
    "uBkQGWQXHRUTEr0OSAsfCD8FEQOoAQgBLAFDAnsElgd0CwMQHxU3GrUeUCKeJAQlDiP7HggZMhHG"
    "B7j90PNj6g/ieNvF1sPTrNJ2087VKtlO3dzhYOZt6unte/Dz8S3yaPHN74zt4uoN6Fzl2+Kn4A/f"
    "XN5B3nPe/N4K4E3hjuIG5Mjlqeea6enrxu4D8sD1J/rn/lUDAgeZCbcK5wntBgUCifsF9ETsL+Vp"
    "30fbVdnu2QXdNuId6VXxu/k6ASoHLQvJDMELaQg/A+P8I/YV8Jjrfelw6hXvtPcDBBUTtCOKNOlD"
    "OFBtWAFc+1rDVVRNzUIcNz8rXCAlF94PfAo2B8UFnAVCBmQHwQjuCa4KKgtpC38LhQvJC1cMBA2s"
    "DVoO6g4VD80OTQ6gDZIMCwtNCWAHIQWZArn/Tfwq+I7z7u5s6gbmEeIO30ndldwr3QPfEOLl5VDq"
    "5u4s85j2E/l5+nX67fhe9nHzifA97kntF+7S8I31XPzWBE8OCBgWIY0oTi32LkAtMyjwH2YVsAne"
    "/aTy5+hD4enbzdjv1wTZqtth397jluju7IzwRvPZ9DL1SfR38uHvg+x/6Dbk3N9+2z/XpNPe0OzO"
    "yM1XzXLN8M34zq7Qr9IA1bbXz9q73YzgXONJ5vjoZOtr7RXvDPAb8C7vSO1g6t/mW+NY4BTe5NxF"
    "3U7fjuLK5s7rI/Hr9bL5Tfyk/Yv9GfzA+Zf2+/KR72btHe0y7zj0hvwiCI8WOycPOWpKzVkAZk1u"
    "FXI7cVVsOWTXWQJOyEE0Nu8rayP5HKAYMRZAFZQVqxb8FzEZRRoBGzUbGxsNGzIbnBtlHI8d+h5f"
    "IJYhWiKDIvIh0yA9H1wdQhspGT0XYhWoE+gRShC0DigNsQtLCvgItweWBpcFugT9A0QDfgKeAbgA"
    "0v/Z/qL9PPy7+kv57PfP9lH2uvZF+BX7Of+qBBQL9BGpGHQevSLcJIIkuSFyHPgUvwvYAfP3qe5d"
    "5pHfc9oD1ybVwNR/1erWnthj2hLcgd1m3qjeLt4E3RHbZNgO1UbRUs1cyXnF0cGjvhu8R7ooud64"
    "b7neuvW8n7/EwkrGHcokzkjSZNZs2kfe3uH25IjnnOlA60/syezR7J7sYexA7IPsau0f77fxIPVh"
    "+SL+IAOzB4EL9w3vDkQONgwHCTAFQQHC/Tn7K/op+7T+/QQaDsYZhid6NmpFIVNmXl1mfmqoaiJn"
    "kmDbV/VNvkPhOQQxjCnLI50f+BydG10b2RucHEsdpR2pHVIduBzvGzgbyxrPGikb2xu9HOIdBx/a"
    "HyUg1h/6HqAd0Bu/GY4XZBVJE0kRlg8yDhcNIAwWC/AJoQg+B+kFowRTA/EBhwA3//n9r/xR++L5"
    "efgZ99b1xfQc9B/06/St9mj5MP36AXsHYQ0oE24YoRxEH/QfbB60GvMUmw00BWL81PMV7JvloeBM"
    "3bzbvtsP3UjfFeIz5Tzoz+qz7L/t7+077aTrYOmt5t/jD+F+3kvcldpc2YnYItgD2BfYYNjE2G/Z"
    "SNqT2x/dAN8I4ULjo+UL6FTqU+zo7fnuY+9G76ruvO3L7AjsjeuQ6z/squ3H70/yI/Uv+Fb7Mv6N"
    "AC4CzAI8AoIAov33+eL14vFI7n/rBupj6vvs8vFu+TwDCg8mHJgpUDZHQZxJzE5+UN1OUkqNQ3E7"
    "zTJ3Kv8itBzZF2MUWhJwEWwRCRINEygUMBUJFqQW5hbzFhgXRhd+Fw8Y5Rj6GSgbTRwoHa4dmh3U"
    "HIwbxhnKF8AV2hMfEsEQxQ8mD9UOxQ7EDswOtw6CDgsORw06DNYKNwlsB3wFfANjARv/v/xh+jX4"
    "V/bh9PLzofMm9Kn1QvjK+0wAlQUZC0cQsxTcF2oZ3xgbFjMRoArJAkP6r/Fo6fbh0Ns4103UCtN/"
    "01HVCNhl2xDfhuJv5aDnxuiw6H3nO+X64UjeZNpv1rzSj88OzVzLoMqVyhPLCsw1zX7Ozs8z0XbS"
    "mdPP1PDVEtdH2KrZ6tow3JHdHt994HzhGOKN4v/iseO+5CTmKei06q7tFfHF9Mr47fzYAPIDMQY0"
    "B70G1wSuAZv9/fhT9C7wGu2v637s4e/29Zv+hwkkFnkjejAiPE1FLkuyTatMkEgiQjQ6WjG4KK4g"
    "mhnnE98PQA3VC1wLlQtYDE0NZQ5hDysQtRAFEU8RtRE7Es0SpBOyFK8VlhaDFz8YsBjOGHwYuBfI"
    "FmcVuRMfEqgQPw9PDsQNiA3MDVQO0Q5wDwUQYRB+EBwQOA8iDtgMLQs9CT0HBAWqAoIAlf4W/Vz8"
    "o/zw/YYAPwThCIAOtRTQGmwgMCWAKPcpiCn7Jm4iLxykFGQMFARc/Kn1LfAb7FvpKOgw6BXpp+rO"
    "7Pnu/PDN8gb0jPRw9H3zrfFx7+js8um35rHj8eCD3pDcA9ve2QvZgtgU2NbXyNe519fXJNiq2HrZ"
    "utpF3BreReBY4kzkNOas5zvoXOg16Hnnfea75Wvl3uUO5/Xo7OvV72n0OPm9/ZcBmwRCBkEGpwR6"
    "ARn9Fvi38pDtVem05hDmfOdX693xsPp8BakRIh7AKa4z0jriPro/Sz1YOJ4x2immIcgZtBIQDdsI"
    "EQZ1BAMEXwQcBSUGVQcRCGwI9Ag+CVIJmAnVCRUK9gpQDKkNHw+uEF0SyRPIFGcVqhVfFZ8UfhMe"
    "EsMQfQ9jDtwN0g0bDvIOIBB8ESQTvhTFFWcWrBb4Fb0UQxNAEYoOowtHCOME7wGG/3D9OPxP/IT9"
    "uv8pA3sHXQy9EVcXXhxPIOMimCMvIqoeJBn5EbcJCQGl+BrxqurM5ZziN+FF4WzipOR351Pq8ew6"
    "79bwsfGG8Vzwx+647O7p5OYR5HvhMt9Q3RvchttE22TbvNsI3FDcfNx93JHctNwD3afdrd4c4Aji"
    "KuS05n3p1OsC7uDvwvAe8SLxlvCt7zjvMe/P7xPxNPOG9sD6Pf8HBJMIYAwWD4gQcBClDlcL+Ab8"
    "AdH8JPhH9LfxFfHg8vT2V/35BWUQ2Bs6J48x7DmjPydCn0FqPh05TTKiKtAikRtcFV4QjAwVCtII"
    "hQjbCIwJKgqsCv8KQQt+C58LhAtmC6oLLgzjDMkNrQ55D0AQzhD2EJMQwg/BDn4NywsWCqQINAco"
    "BqcFlAUHBtcG9QeuCXoLOQ0DD2oQSRGzEVoRdRBPD1wNuwrdB8cEpwHN/hr8/PnU+Jb4V/lX+2r+"
    "eQJYB8cMQBILFwUbdx0QHuUcxxmSFCEOwwbT/hT3AfDP6SnlM+LA4K7gwuHa46jmmulH7IzuJvAQ"
    "8R3xafAa7xHtt+pq6CPmuOOp4fPfmN6r3R7dr9x73KPc2NzZ3OTcAt0N3VzdEN7t3hXgqeFK4wXl"
    "Fued6JjpVep86uTpVOmY6L7nkecS6O/owuq97UPxTPXB+Qz+8AE6BaIH3wisCEMH1AR+Ac/9S/ry"
    "9pf0wPNX9OP2vPtgApEKThSdHmIo6zCdN6I7vjxaO+A3nTJaLKAlAB/9GLgTVA8dDBcKGQkDCYsJ"
    "aAp4C1MMUQ0bDngO0w4mD1oP3g+jEDkRJBIwE9gTdBTYFNkUgBTuE/ASpBEtEMEOVw3RC5UKxwlR"
    "CTgJewkYCigLZwzJDQ8P5A9sEJoQ/A/dDmYNIwtiCKwFxQLN/2/9p/ta+u75mfoJ/G7+1QHABTgK"
    "JA+6E3MXShq7Gy0b9RhhFUsQEgppA7n8IfZh8MvrKOij5VnkPOQd5bPmgeiY6rDsh+7h79HwKvH7"
    "8DXw9e5D7SfrAOnS5onkTuJf4KjeHN3/203b1dqn2uzaJttz213cbt1K3qrfUOHt4q/kneYz6ETp"
    "7elO6jbqkuny6JToR+hv6FHp4epN7ZrwgPTl+GX9YQHkBIgHzgimCDoHnQQmATv9OPlz9WTyrPDC"
    "8LTymvaS/JgE9A07GJUi/SuIM9U4MTuwOs431DKCLOYlLx/eGOoTLhBaDc8LbAvBC3gMbw1cDiQP"
    "wg9JEIsQpBCoEKcQ2RBkEQEStBKcE50UaBUXFr4WERf5FsAWNxYVFeETjhLmEGkPSg4gDRwMfQsf"
    "C98KygrSCvgKIgsCC5IK/QkcCcUHAwYlBAACrP+N/YD7ivnp9/r20fZl9+/4efvg/g4D2Ad+DK4Q"
    "WxSMFrAWbxXDEhwOWwguAnz79vR377PqzOZk5ALjgeJd49jkS+Y/6HTq6+va7JPto+0J7S/s6+oT"
    "6Tbnr+UU5K7i9uGF4UPhleED4i7ieeKS4lfiQOId4tPh5OGV4j3jA+Qj5XXmr+cV6W7qXusP7LDs"
    "LO2n7QHuRu7Y7r/v8/C38gL1tPc/+/7+rwJNBlYJSAsyDKELvQnyBmQDUP97+xv4ZPUp9NX0N/dB"
    "+0wBGgnfEfcasCNeK0UxwzSiNUU09TDcK/Elyx+dGeQTZQ8ODPMJ0gioCC4JAQq4ChELRQtrCzQL"
    "ygpwCv4JkAm4CVYK9QrOC/0MRA44D80PCRDZD1sPlg7HDeMM8wstC60KZAo2Cm0K+QrSC5gMYg0m"
    "DrsO9A7CDkcOnQ2gDGYLHAqUCMYGFwVXA2sBt/8y/uv8d/zD/Kb9hP8yAoMFggnXDcsRIBWNF8kY"
    "eRiCFhcTRQ6DCHsC9ftg9cjvQOuO5ybl6uPE47XkUOY06Hrq4uz37qTwCfLr8rzyv/GJ8LfuTuz/"
    "6dfnpuWX49Xhj+C73/vegN6q3h/fl9824AXhqeFA4t/ikeMj5LTkf+VI5gHngufI5yToc+iZ6PTo"
    "p+lQ6lHrJe1+7y/yg/U++Rb9xwAXBNQGrwhQCSIJQghSBrMD6gDl/dD6SPjG9rL2H/gk+/v/XwbJ"
    "DeAVHR6GJXAroy/qMRkyLjC5LCcoAyOpHXcYthMREGwNlwvBCtQKaAtUDHYNcw49D+IPVxBjECkQ"
    "6A/GD+QPKhCPEDwROxIHE8YTbBSPFEIUpBOaEg0RQg9YDbULXApXCbAIpggjCfIJ8Qr+C9oMVA3P"
    "De0NYA13DFULuwkBCFkGlATfAncBVACZ/2j/vP+IAAkCPAThBuoJUQ3KENsTuhYJGRca2xlZGGAV"
    "AhHGC94Fjf+K+Tf0su9G7NzpFugv5zbnj+dM6JPpv+ra62Xtqu4U7//udu5n7SLsluqz6B/nrOUB"
    "5IriK+HA37Dewt3U3HLcdNyB3OPcl91I3jPfZ+Cd4f/iiOQf5rbnG+kf6uTqb+uQ65LrmOu66wns"
    "6ex57nXw+vIc9nf55/wsAOUC7gQpBjIGUQWrAyAB9v2u+mL3e/SJ8rPxXPLS9PP4bv47Bf8MKxUq"
    "HUkktSksLcEuSC65K64n2yIiHT0X5RFhDXIJrwY4BYcElwQzBREGCwcLCNkIogllCvcKfAszDBcN"
    "Cg5TD9kQsRKpFLcWqBhwGqEbDBzlGxYbgxmpFywW9xQcFOITRxT1FAkWeRfdGO0ZyBpUGz8bcBr1"
    "GLIW2hPZEHkNzgkzBlsCf/75+p/3M/Q78RPv2+327W7vYPIz97z9UAWHDa8VmxydIVIkVSRpIekb"
    "tBRrDKIDU/sk9AruEemX5Vrj6eEh4b/gn+B54AngL9/u3SLcq9nS1qPTN9CLzADJzsXfwm/A7r5H"
    "viK+5b6fwOfCgcWryE7MANDn0+zX1tvI37njkucp64jucfE+9If3Qfu2/6gF4QwbFTYehCfeL3A2"
    "ezqnO0s6aTZyMHApGiLvGn8U6w6TCq0H+AVPBb8FrQbdB6gJ2gscDugQSRT5Fz8cPyF6JnorzC9p"
    "Mw82uTdnOBs4wzbbNG0yWy8WLNcobiVSIh0gvx4xHiQfTSFGJLonGSt/LVAu8CzsKGwi0RnXD3sF"
    "hPtT8nrqd+QH4B3do9tl26DbMtwS3f/dit7M3sjeQ95x3ZPctdvp2nvaCtra2WnaT9t73HPeIOEU"
    "5KDn5OsV8Dz0LfjD+97+awGAA4oFgQc8CQQLBA0DDxwRdRMoFlcZvByHID0lhiq/L5c0cDieOh07"
    "oTkdNhUx5Sr0I4Qc6BRFDQkGjf/g+Qr1N/F77mTsqOp46ZTofudq5oLlxeRP5PDjv+P341rkoOTs"
    "5O7kUeRN4/rhEODg3Y/bT9mU15HWjdaj15DZLtyP30vjB+fG6vDtZ/AY8rDyH/LX8MruaOxP6mTo"
    "l+b75D/jC+F63vvaddZd0Z/LcsWRv0K6lbVYsvawNLGOsye4mL6XxiTQjNo15Ynv3fioAKsG9gqb"
    "De8Oiw/wD2IQXxFIE7MVlRgMHFAfriGrI/YkUiVYJXUluSXZJoIpfi08M4c6dUIkStNQflWfVwpX"
    "01ONTh5IREHwOlI1wDCZLeQrOitrK74r2Su/KxArYilSJ8EkdyEuHhcbARg4FSsT3xE4ESMRehFT"
    "EkkT3RP9E5kTlRIqEaMPag7VDfwNLQ9gEZQUgxj7HGkhQSUUKEkpbShEJe0fnxgCEL8Gw/2k9cru"
    "aulc5XziqOA+3+vdr9wV2+7YU9Yo003PScsWx3jCyb10uXO1JrKgr6yta6warJKs5q0EsJqy0rW/"
    "ufy9gcKMx6HMX9EN1oLaZd4D4krl7edL6uvsfO9A8hD2zvqHAI4Hpw9NGEUhmikyMNg0UTdDN101"
    "LjL8LUMpfySrHzEbOBeMEx0QBQ0LCnEHvQXiBNgEAQaeCFoMLBGwFk8cUCFsJTAoDSkGKI8lGSKL"
    "HV8YNRNlDkYKTgdsBaMEJAW/BvoI5AsdDxQSqBSTFmgXMRf4FV0TVg99CisFxf+7+lP2nfKU71Ht"
    "kesr6v7ov+cn5nDk6OJg4ezf6d5j3jvem95k32zg0OFS45XkpeXL5rrnbuiB6d3qQuwM7nDwGPMz"
    "9uX5sv2FAcUFKgpCDigSrRXDGKYbaB4EIeUjMiffKjovQTTYOaQ/DEVfSTBM1kwpSxhH80CmOb8x"
    "iCnhIXgbJBYxEtoPbg6+Da0Nlg0QDSoMkAodCCwFtwH7/V36zPZu87Twou4M7QLsVOvM6mDqx+mF"
    "6L3mVORp4TDeVtvi2AbXT9bW1jLYetqJ3bng3eO55ojoFump6CznmOTA4d/eXdyf2rjZrNlB2gPb"
    "0tuY3NPcctyJ2+nZx9ek1ZjTvdGs0JfQndEa1PDX5Nzh4p7pdvAj90L9YQKgBtQJ5AtEDRoOfw7K"
    "DmkPRxBjEeESlBRUFvgXXxm9GgAcXh1DH+YhkiXFKlox9ziPQT9K3VEFWARc+FxBW4VXC1KHS61E"
    "9j36N/Qypi5yK3Ap5SfNJhEmPiUlJNUiKCErH+EcSBqRFwAVbRIlECcORwy0CnYJbQiZB90G9AUd"
    "BVUEiwMBA/ACHQP+A40FsQd6CqsNgBDwEpIUwhRCEzYQQgvHBKL9Evai7h7o4+LO3vTbDNq22AXY"
    "odcK103WhtUk1DbSVtA/zgDMAspWyPzGQMY+xpPGQMclyCrJesrny0zNKs9X0afTcNbu2aXdsuH4"
    "5SDqIu7a8dr0GPeH+Ar5Hfn1+MX4Z/l3+7L+TQO9CXURnhkHIqYpWC/qMlE0ZTOYMKcspScBIpMc"
    "tBcyE0UPLQzlCWsIjAc1B3cHbAjZCZkL3w2BEDIT+xWtGLkaAxy2HKQcpxsSGswX0BRjEcANIQqj"
    "BlQDsADt/hX+Of53/5gBSwRLBwsKcgwHDlQORQ3zChQHMwK2/Ab3zPFZ7Y7p8eaA5bnklOQF5W3l"
    "yeUn5jPmAuaa5fDkDeQi4yLiT+GM4MnfIt+q3mLeV96P3iLfMeCC4R7jPuWt51/qbu288CP08ffR"
    "+5j/TAPWBvAJzgx0D90RTBT6FvgZZR2JISMmOiuYMGU1EDlkO6c7gjl0NcIv4iivIY8a+xOWDoIK"
    "bgeVBbIERQQmBAMEeQNqAssAo/4b/Fn5afar83zxwu+Y7vzt8O0e7ofuGO+F73vvE+977rDtyewa"
    "7NPrIuwJ7bru7/Bm8yz2DPld+yb9h/7P/tb9Mvzr+QP34/P38IfuzeyO657q6ukg6fTniubC5Gzi"
    "299C3c/a59i01zTXsNdp2SLcBeDZ5AbqQO9w9DP5/PwUAFQCugOgBFAFxAUiBuEG9gdMCe4KtAw0"
    "DmkPdhAsEY4R2BFgEoMTiBWBGLIcOCLHKOgvLTewPZtCcUXjReFDuj8LOq8zCS2SJgEhmRxCGUIX"
    "dBZcFqwWchcdGH4YZRjsF8kWAxXOEmQQyQ1OCw8J3gbgBFcDzwFbADX/KP7l/Jn7dPo8+T342/cs"
    "+O/4k/oe/RcAWAOgBloJMwsbDJMLcAnhBcsAwvpO9Nntyues4qDeudub2VnY2tex14vXmNeK1wrX"
    "U9ZS1eHTPNJy0IbOt8w1yyLKqsmGyb3JdsrAy2nNcs/y0R3Vbdji263fqONk5yfr7u5m8p/1cvis"
    "+mT80P35/igAnwHWAxQHSguWEDcXnx5gJgAulDR/OZ08dD0tPEU5IjURMJgqKyUdIJkbnRdSFLER"
    "lg80DlwNvAy3DD4N7Q3WDgYQSBFtEoITdRQQFUMVCRVEFAQTWhFWD/UMbwoWCNQF2AM0AkwBLgG4"
    "AdcCtwQEB1EJsgvEDe0OOg+mDskMogmoBSsBcPzY993znfAU7lDsJ+sz6onpBull6LPn+eYv5ojl"
    "7OQP5H/jXONc43nj9OOG5DflAual5jLn3ueE6FrppupH7DXul/B186v2GfqL/fUATgRdB/UJEQy3"
    "DQUPKxBfEdQS2hSWFxMbfh/BJH8qBDC9NPA3TTmNOKI1tzCAKn4jfhwtFqwQRAw/CZYHnwYWBuMF"
    "8wXGBegEpAMBAtL/Jv2T+gj4vPXx86jyuPFX8VbxM/Hz8JLw6O+37jPtp+s76uzo+ufD513otenI"
    "647ul/HJ9PL3pvp8/FX9Nv3G+zX57PVp8sPuZeun6NrmAeax5cblI+Zt5jzmr+Wo5BfjfOHn31Pe"
    "Fd1+3LDcwN3B34TiF+YL6sDtRvG39F33WPkO+2H8af20/jkA3gHoA1wG6Ah+C+wNLxAaEnoTWhQ0"
    "FTAWFxd+GO4abB4JI8wocy+KNlA92EKjRq5IpEhLRltCaT2wN9kxeSyUJ3MjcCBaHr4crRskG/oa"
    "rxoxGnoZTRiQFoUUNxLED1cNBQv3CEEHmAUOBMUCRAFw/7/9A/wS+lP4CPce9gr2v/Yj+Cb6wfyR"
    "/2oCDwUKB+oHnAcBBgIDnv40+SDz3Ozn5n3h7dxm2ejWcNWj1D/URtR81FDUGNTa00/Tl9LE0aDQ"
    "ac9gzjLNOczDy63Lx8s5zBzNTM64z8HRkdSC17/a3d5z4/Hn3uzM8TT2LvqS/T8AIAJDAwAEuwSb"
    "BesGRQnhDPMRfBgBICwoVzCsN589jkEOQ5tCNEAOPAg3iDG3K10mrSGZHY8aXBhrFuoUChReE9wS"
    "fxI8EhAS3BHKEfQRERIFEiESMhLhERYR/g99Dl8M6gk/B1IEUwH1/hT9wftW+9P78fyT/m0AfwKg"
    "BFcGcQcZCOYHtAa6BN0BUf5b+nf2D/Mo8JztAew566XqOuob6uHpYOmJ6J7n1Obl5f7kVOSr4//i"
    "veK54oziouL54lTjjuPL4ybk1eTz5ann0emM7CbwPfSU+Gb9NQJuBkkKnA0lEDQSsBOxFMAV9hZa"
    "GG8aXx3wIFklWiqHL1w0+Tf+OTY6cTjwNDMwaSozJE0eBhlKFGgQZg1fCwUK5QjzBzgHSAb0BDcD"
    "LgHb/nT8RvpL+Kz2nPUp9fH0svSA9E30cPMx8urwmO/A7V3swutn61/rROzI7Z7vx/E09Mr2Tfmp"
    "+8b9K/+L/w3/Zv2e+gD35vKt7r/qM+d35OTiBeKe4bjh5eHj4bjhLuEd4NPect1q3Ajc/duD3N7d"
    "tN/h4WLk7uZY6c7rIO5J8Gvyi/Sq9vX4dvsX/tEA0AO0BlMJ/AtkDj0QnRG5El4TuhMVFJYUoRWh"
    "F+8acB/bJPYqZDEnN6I7SD7UPo09fzruNdMwnitsJtUhDR4BG8AYNBc9FvgVyxVrFUUV7xT3E8oS"
    "WhF2D60N9AsRClAIzwZbBRME2wKlAY8Aif9w/nP9q/wp/L37dvu3+3P8df01/2sBqAMKBoQILQrj"
    "CqEKFAlDBmUCpf05+KDyP+1n6DzkCuG/3h/dHdyF2wvbndot2qXZIdl22KTXz9bP1anUwNP80l3S"
    "LNJz0tjSqdP21JPWXtiq2kHdMuCq44bnwutG8Of0Jfkj/dwA7gNJBikIcwlxCpMLzwxZDvsQohRW"
    "GQUfcSUiLLoyWThXPIw+oT6XPP84PzSTLpAoxSJjHbIY7RT9EdAPVg59DRgNBg1IDYgNkw1uDQIN"
    "ZAy0C9EKnwl2CFAHAgaIBPECKQEs/+T8dvof+CT2Y/QT82jyR/LM8vjzpPV696L5KvyI/kMAhQEW"
    "AkoBHf9C/Lf4b/RK8NXs/ukC6A/nwuYD54/nLuit6OTozuij6DXolufk5gjmGuWF5D/kHOQ55I3k"
    "HeUP5jbnbujY6VLrHO1973zyuvVs+Wz9cwFbBRwJiQxyD+0RzhMqFUQWFhfqFxAZxxooHXIgqiSY"
    "KRMvpDRzOR49JD8SPzE9vjm/NNEuoShsIrocFhhbFKMR/g8cD5oOeA5ODtUNAg2pC+UJ/AfMBX0D"
    "jQEiANH+u/1K/Rz95fyl/Fv8tPvN+sb5l/hg9472Avax9fj1Bfe++Af74f3rANQDPAa5BysIWwdM"
    "BSkCNv6++Tb1+fBY7WXqJOiR5prlweTV487igeHW3+LdxNu82djXVtaC1UHVdtVU1tjXttnU2+rd"
    "FOAp4unjceUH553oeOoE7c/vxvIq9vX5ov0bASUEygbPCFgKiQtuDEENcA4xEKMSOxbJGjEgbSbP"
    "LJQywTeGOyw97TzBOsY2sjH7KxMmqSD2GzAYpBUUFDwTPROOE/cTSBQ1FHgTVhKVEC0OYAujCCYG"
    "BwRIAg0BKgB7/xX/tf4Y/kX9J/y2+l/5IPgS91X2Ovbf9iD4wvnw+3X+sgBFAiUDEgPLAVP/1Pt4"
    "92XyHu0g6Kzj2t/r3LnaS9ls2OPXjtcn14HWitVI1JjSCtGHz/7NtMzxy2TLMctZy8DLf8yQzdTO"
    "hdCn0iLV+9dv237fAeTb6Dju2vOm+ZL/mwUZCxUQuxSdGOwb7R69IXkkqydyKx8woTXPO6RCnkms"
    "T0ZUHVeLV5VVTFH1SiFDhTqGMdwoGSGOGr0VkBJqEC0Paw5qDbQLXQkXBioC4P10+Q/1zvCi7L3o"
    "A+U24YzdCtps1g3TA9AdzbfKbMk6yfLJgcswziHSpdaM21rhnOdK7Vvyr/bK+b/78vy+/dj+vwDh"
    "A9cIUg+UFgMeiSQfKY0rwit8KUYl7h+0GfcSPgwqBqAAyPtl+F72SPU09Qb2AvcN+ND4kPhU90H1"
    "PvJ+7ujqyefT5Y3lKOeX6hTwbfcMAI8JRBP8G04j7ig8LCMtdSzPKu4o2SflJwMpqCv+LhIynTRR"
    "Nj02UTSxMFsruiTDHRMXHxFpDDEJYAehBpgGtQaaBicG8QRbAuT++/qO9mfySe9d7U7t1e+79Or7"
    "+wR3DiQXMx5UIhMj6yBHHC0W+g+uCvUGfQVtBtQJyA+zF7EgwyktMtw4vzxvPbg6bTQ7KyQg3hMp"
    "B2j7YvE26XfjOeA53xbgQOIe5TPoseoH7PvrBeri5SbgpNnL0kHMy8bGwiXAs74Cvg++Bb+hwLXC"
    "ZMXuyP3MedFV1knb59/z42PnKeoM7D3tQ+6k713xq/O19sn64f8HBmsNTRaFIL4rxjchRBRQtFoi"
    "Y7Zo0moeaS5k5FwAVNpKsULcO4c26zKDML4uSS2cKxgpqiVcIVcc4xZhEUAM7AdPBJgBpP8h/sT8"
    "TfuN+W33yvSU8dztzenf5Z3iLOCM3t/d1N0V3pXeF99B3xDfeN5L3VfbDtg/03LNxcZDv3+3fLBj"
    "qkelxKE6oP+f4qAOowqmQam9rHqwYLS4uFC9/cHoxrfLF9AW1LbX7NoY3ozht+VJ61vy6vqwBBwP"
    "MhkEIuQony1BMAMxvDC4LwIuLixjKmEomyY7JRUkfSPdI2QlaSgjLSgzODqPQRVI/ky0T4lPs0yr"
    "R7BAmThiMIgouCGIHOYYyBbVFakVhRW+FP4S5w91C/IFCACW+e7ykOy85k7hgdx82FfV+tIY0Z3P"
    "os7wzSLNh8xUzHnMDs2FzhrRxtSA2ULfC+ZG7Xn0T/s3AfIFmgnqCy4NIQ40D48QExPmFlMbth+F"
    "Ix4mKCevJvwknCLOH7AcrBnjFmMUPhKXEEcPLQ48DXsM9QvAC94LPAx4DJwMjgzzC70KSQm0BzEG"
    "KQXgBGgFDgd0CU0MdA9uEo8UqBVRFVAT1g89CzEGhQG//Tj7RfrB+iL8/P3H/64AGwDX/c/5QfSj"
    "7bLm1t+t2Z7UHdEHz4zOmM/q0VPVkdkV3l/iWebF6arsFu8S8UbzNvb/+Qr/mwUjDfoUiRyiIocm"
    "3CeKJk0jvh4uGYoTlg6HCrYHfAbEBkAIjQo8DQIQZxIzFD0VWhWCFJgSmA+2Cx8H2AFf/Or2s/FR"
    "7QLqd+cu5hDmceY154vooOkJ6rjpfOho5i/k1+GK3+Xdydys28faHNpP2XbYwNdG1yDXudcN2Sfb"
    "/d2G4WHlX+mU7ajxU/X++Iz86/9MA64GHwpCDugSABgbHhwlDS3qNV4/I0ngUqxblmIkZ7hoA2dj"
    "YtRbTFSQTHZFjz8AO6o3jTUzNCQzAjJfMEcupCt1KDklKiIvH1scGBrvF4oVDRN/EGYN3wlCBqcC"
    "TP95/GP6Bfmf+Pv4zPnI+pv7yvv3+i75oPZx86rvT+tk5hbhhtun1drPK8rAxP6/BbytuDC2oLTZ"
    "s9yzb7RAtY62V7g2upi8pr8nwyDH7MsN0TzWwNtz4Rvn6+xE8z365wFmCpcTFR1MJmkuuTQpOfc7"
    "Pz2EPXg97jzuOzk7uTobOsU5njl6Ocg5jDqRO4Q9SUCFQwZHPkqHTIJNp0yjSbVE5D3WNVwtCyW4"
    "HfMXkRPMELcPaw+rDyAQfA9eDX4KegYsAU37D/WO7g/o0OHR23LWtNF6zfjJ7cYTxMfBLsD8vl6+"
    "fr5lv+bAPcOExrjK1c/O1T7cy+I36fLu1/Pt9zj7uP0eAOkCMAYfChwPmRTdGXAegSGoImAiwSDh"
    "HbkaiBfjE34QjQ2JCuoH+wUlBFIC8gDv/x//iP4x/hP+7f2x/T/9f/yh+5L6Qfk2+Lr3tfem+Lf6"
    "fv0RASUF+whGDJwOPA8hDt0LmgjuBMkBh//7/UX9KP0H/df8GPwm+jT3RPNL7tPof+Ny3jjaL9dO"
    "1YnU9tSv1pfZKd104VXmROvW7/zznPee+if9wv/fAp8GcgvHESEZ8yBuKJouwzKLNJwzyDC+LAko"
    "liOZH1ccVxprGU0ZQBqMG7Ac4R2EHlweuh1WHCEakBclFNoPMAsPBmwAtPok9frvheu65+rkNeMT"
    "4pDhgOFB4aHght9V3WnaLdeX0zTQcM2MyurHrcURw1TA7r1nu+u4HbfBtW21S7YYuEW7wb+0xE/K"
    "UdAe1tnbmuEP56rsM/KF90n9cgPbCfkQyRgxIW4qWzSjPgBJ1FJMW6RhTmXfZV9jtF6YWPZRf0u7"
    "Rd5AXj3FOtU4yjfkNiw1/TIyMHwsYSheJCYgKxy5GIQVdxJmD18MWQkyBvwC3v8X/eH6VfmW+Mj4"
    "cfki+uD6VPue+uH4UfbR8rHuaerR5bDgcNsL1nXQ4MplxTfAZ7v0th6zHbDPrTqsravcq6isGa71"
    "r0ayMLWxuMe8VsEaxmfLCdGq1o/c8eKH6XDwOfjqAGkKnhTQHlUomDDyNg87Tj0+Phs+jD30PFI8"
    "xzuSO3I7nTvpOxY8RDzBPHs9wj7CQAdDXkVbR4xIjEgZR/lDjj/wOZkzFi3DJhkhpxxpGWAXrxah"
    "FswWyhbhFY0TIxB6C4gFDv96+LTxCOvH5NnedNmP1B/Q5Mv3x0fE+sAkvva7m7rxuS+6cbtivTLA"
    "BsSgyN/NudPH2bjfjOW86lnvkfND98v6wv5TA8UIOg/3FUscryEhJWsmwCVlI9IfgxvSFmUSOw5G"
    "CvAGPATYAf7/ZP4S/R38nftZ+377yPv2+wv8tfvg+rH5PPi09lT1ZPRD9A/1svZe+d/8zADiBK4I"
    "ggsXDRENpAsfCdwFXgIu/4z8Rvqk+Fr3A/Zg9HDy2O957K3oluRl4LfcydmY13jWYdZS1yDZ8Nue"
    "3wDkveh/7SPyTPYJ+mv9egCPAzMHwQtwEWEYWSDzKGExUzgoPbI/gz8hPXE52zTxL2krbCcfJMkh"
    "PyBBH7wegB5PHgQeeh2sHLgbWRpYGLsVbxJmDrcJuQSK/2b62fUH8vvu++wL7NXrOuwI7dTtFO5y"
    "7cbrBOlJ5QThltwW2A7Ud9AJzQvKhscRxczC4sA3v829srwqvIS8nb2ov7jCWMZ9yorP3NR62qDg"
    "/OYl7VXzbflu/4IF6gu1EikaVyI7K6M0UD7lR31QFldEW71cUluTV5tS4kzqRppBAj0KOR42+DMS"
    "MmEwgC5ELLwpoSZtI3AgWB1sGhEYZhV7EqMPUwyuCA4FPgHL/RH7w/ht9yD3X/dH+I35x/p++zz7"
    "//nX93T0efBX7J/ni+LR3fvYB9SNz2DLU8eWw17Afr0LuzW5KrjMt/q3zrgxuhS8hb6TwVbFuMl9"
    "zmDTYNh33X/iqOc67SDzefnDAPUIAhKzGyclBy6oNf46KD6HPyI/8D2VPOQ6VzlwOLU3QTcIN/Y2"
    "yDZzNjk2QjZ0Nis3ajidOaU6QDv/Oq45Czf6Mj8uFimEIyMebhmSFbASxhCpDzMP0g5ADtIMWgre"
    "Bi4Cp/za9pHwIuow5HfeINl61EbQhsw5ye/FyML+v4O9rrs4ulm5h7lZuqG7Mb6/wf7FYctl0c3X"
    "cN6X5FXqBPAn9fL5vv6IA7EIrQ7OFAobASG+JdMo6ykYKRon9iM8II0cNRkeFncTDRFNDwMOmwxd"
    "C3gKcwnBCIwIUAgfCCAIxQf4BqAFAwRCAkQAc/5c/e38P/2S/pIARgN8BrMJjQyeDm0Pzw7lDMoJ"
    "5AWZAXf9ifnu9QDzq/CB7qXs7+rE6CPmYuM94Nzcv9lE1z3V4dN50/DTLtVv19zaBd+m457op+1f"
    "8pX2VfrZ/XgBfAUJCmwPBBaQHW8lNC0DNN44ZTvQOzg66DbFMo8uXiptJnMjQiHJHxMf/R71Hgcf"
    "Ih8UH6ce9R0XHZcbYRmvFkgTLg/ZCk4GnwF1/cf5sPab9FDzuvLt8oPzB/Q69J/zJfKd7yLsQOgz"
    "5OnfAdxC2JbUO9EgzgrLGcgBxS3C678JvsS8qry0vcW/xcJpxvfK7s8T1dLa3+DZ5vnsHfP0+Mr+"
    "sgS0CgsR0Rd0H6wnUjB9OZJCk0oFUVhV1lanVWBSuU00SGtCKz2iOJk0fzFjL6wtRCzrKmEprye+"
    "JX4jTCE6Hx8dwxo0GKEVCRP5D64MhAl2BnYDugCn/ln9z/zT/JX9uv7J/zAAuf9r/gv8w/jY9Jfw"
    "H+yF59nict5R2hvW+NEAzgnKRMbSwvW/0b1hvMS7DLz4vGy+isD3wgjGoMlwzbPRMday2m/fIOS7"
    "6MPtDPOD+Bf/VwYFDn0WAB+cJs8sNTF5M7MzgTKIMPItMCvEKIEmnSRcI0IiUCGmIPEfRx/OHpQe"
    "8x5sHwUg4yA4IcEg5x8zHm0bSRi/FNcQ/gyuCUoHrwXkBD4F/gWuBkoHvgboBG0Csf7Y+c70b+/q"
    "6fXkZOBK3NbY1tUP02TQ3c2gy2jJmMeVxsfFf8VFxrzH0sm/zJTQVdW52sngOOen7Qn0Yfrn/w0F"
    "NgoWDwUUeRmIHywm2iz+Mj04ZTuBPPk7bDmpNbEx7Cw+KJMkCiHUHa4b+hlzGDcX+RXnFLETmRKY"
    "ETcQwQ4PDWwKdgewBFsBLv6z+3P56/do96n30fiv+gv9sP+/AQcDWwMtAuL/vvyZ+EX0FfDk62To"
    "fuUY4z/hWt9p3Xvb8NgV1mTTndAzzo3MTMvbykvLdsx5zkjRxNT22FDd8eH15n7rx+8N9AP4IfzZ"
    "AAgGLAy3EwgcviRGLZw06jnPPEQ9hzs0OCc0ii8DK2EngCQ4IhMhwSDJIDchoSHmIR4iDyKiIQ8h"
    "8B8tHqwblRg5FWERNg0mCX8FEQIe/8L8Wft9+ij6cfq9+oT63flF+KP1dvKb7jLq0+Wj4YTdutlk"
    "1njTj9CczdnKMsigxcPDt8KRwnPDUcU2yAXMXdBV1cXaceBh5nnsoPLT+KH+JgTVCaIPRBUuG7Ih"
    "FSkqMak5MUIxSvVQsFWwV1VXIFUdUTJMNEcoQm09YTnlNfQyizBmLlIsCirHJ7wlgSNVIVwfRx1P"
    "G1sZ8xZVFKgRsg6fC7cI1QVAA/YASf9f/un91v0s/m/+G/4d/UH7dPjz9Pzwd+y05z3jFd/Y2jHX"
    "09NA0LDMX8kUxijDj8CRvom9O71xvX2+WMDZwt7FOslLzfPRl9aG26XgmuWm6r/vyvQv+vL/BwbZ"
    "DDgU4Bt+I1EqtS9UMwU1IjXrM7cxEy9aLIkpICf9JPsiOiHNH3seJB3xG/waIhp4GTAZDBncGHMY"
    "jhcXFhIUYxFgDjML+wcLBYkCfQA//7X+ov7u/kn/Lv+J/uz8WPoP99vy1O2p6KXjed6w2XfVsdEY"
    "zr3Ks8e2xL7BKL8KvTC74LlNuWO5ILrIu0u+n8HWxZ/Kxs871fXauuBP5s3rePE+9x/9eANBCoER"
    "MRmoIFon8izdMNYyEDPWMbwvMi2QKt4neSVlI6whHiDHHqMdoRzQGw0bexouGgca1Rm0GSQZ9Rdk"
    "FkMU3BF8D0gNjQuEChQKUAo5C5IMPg6+D+8QrRE6EaMPVQ3HCUQFfgBk+0b2q/FX7W7pP+ZX43Hg"
    "Zt1H2k7XOdQa0YfOd8wFy1jKWMp0y4zNLtC20+jXKdzC4ITl3ekr7mPymPZT+2kAyQU5DMQTsRuj"
    "IxsriTHuNSY4VjhiNgozRi8CK9AmNSMeIJYd3xuhGsQZNxnGGI8YLRjBF78XYBd9FpcVKxQgErAP"
    "2wwMCmkHvwSyAjkBCQCR/5f/1/98ACEBUQEaAfb/5f09+9v3z/N77ybr7ubL4vLej9s22OPUqNFv"
    "znHLAsn9xrfFa8X8xYLHxsnZzOTQI9W/2RXfc+S+6SDvKPQV+Rn+2QLCBz8NChOWGbkgSShAMNI3"
    "KD4gQ+1FaEaHRV5D8z8cPD44mDRtMbMuTix5Khsp3CexJq8lyyTCI8IixCGuIH4fCh5HHEoaBhiY"
    "FToT2RCLDngMZQrMCLcH1AZHBgEGigWNBNwCWgBA/SL5WPRF7+bpkuSc38LaE9bh0d3N08nexSPC"
    "v76Su/24Yrd7tkS26rZOuG+6Qb18wFDEp8gNzcHRqdag25LgvuXp6nHwjfb8/MgDHwvmEkoa4SBp"
    "Jn8qBy0sLjguey06LMMqYSkHKKImjyXJJO0jGyOIIjsiAiLcIdQhGiJGIjsi1CHlIIQfnR04G5sY"
    "yBXzEkoQvw2ZCx0KNQnZCMgIzQicCOoHhgZLBCIBSf3d+Bb0Xe/o6pnm5OLF3//coNpi2CbWENQ0"
    "0oDQXc9jzu3NW85yzxbRfdOY1ivaG9534g/nXuuX7xn0Wfg+/FcAmQT4COcNaxMCGU0e5SKmJjMp"
    "MSqlKfEnhCW6IqAfkxzVGVoXChUvE58RRBAnD0cOeA3oDJgMYQwbDNkLMgsFCpYI2gbvBBMDWAHC"
    "/6f+Af7//XX+P/+KACQCPwMNBEwETwOMATz/+vs2+IX0y/Bh7XTq7ecY5m/kr+Iw4ZPfv90t3JPa"
    "4Njp14vXndd92N7Z8Nv03mvi9+Xq6S3uTvJA9gj6v/0wAZYEYAi4DIIR2xa5HLQiUCglLaQwVjKn"
    "Mssxqi/YLNMpxiYgJPkhDiC5HucdSh3yHJYcJRzcG2YbrBoHGgAZcxewFZkT0RDlDdkKuQegBIEB"
    "l/4o/A36afgW99b1z/Ss8/vxuu/47H7pl+Vq4Qvd49gg1azRgs7wy9jJ3MccxrHESsMuwqrBicEO"
    "wk7DKMWgx+XKxs4d06/Xvtwq4nDnrOwu8lT3QvxXARMGyQoOEIMVQRtBISYnwyyyMT81hTeDOAs4"
    "pzZhNKMx9y5DLLkptifTJSkkFyMKItsg1h/BHm4dEBy7GlMZvRfVFQYUEhLfD4oNOwv6COsGAAU2"
    "A7oBrAD+/7H/l/9L/5r+af3W+5356var8yjwqOxL6fTl6uIP4HfdJNvi2LfWs9S20t/Qf898zgfO"
    "M87BzuLPk9G903LWitmm3DDg8eOu52frNO8i8zv3afvA/6wExwkID2QUdBnyHcshlCRNJg0nByec"
    "Ju4lJyVtJNwjVyPcImYi/yGfISUhmCDiHxsffR4EHnod9hxcHJ0bpRpLGbUX9hUWFCISJhAeDoYM"
    "PAspCoMJSgkuCSMJoQh9BwcGFgSTAdb+0/uO+IL1pvIN8NXt7+sb6nTo0uY75bXjP+IL4Tzgxt+r"
    "3yHgD+GS4pfkIucB6kPtlfDi81r3mvqL/awA+AMYB2cKIw4dEjcWSRrpHRAhPCMpJDkkWCOpIagf"
    "lB1bG0UZOxdaFc8TNBKkEFsPDQ6wDG4LMQoJCe8HuwZmBeMDDQL8/8b9uvvY+fD3SfYb9VP08fPx"
    "8zP0w/R39dL1s/X89P/zcvI58P3t7et36R/nWOXN43DibuFd4GXfWt4D3cLbi9p42bfYTdg62O7Y"
    "9dll27rdlOCk4wPneerL7SrxSfQ/9y36Hf0KAFkDGQcZC1APxBP+F7QbuB6AIBIh+yA4ILUe9hxF"
    "G6YZGRitFpwVzBTsE00TxhL5EVYR8RAoEFwPrQ6DDTQM6gomCRsHOAU6Ax0BNv9+/c/7iPqt+SH5"
    "wfiN+Cb4evdg9s/0+PLQ8EHuuutK6eDmrOSa4qXg8N5I3Znb+tkc2ErW3tSm07zStNIa0+LTZ9Ve"
    "18zZ0dz53xrjiOb46S3tcPDL8x33nvpW/mkC6QawC44QcBUwGl0eviEEJFcl6CWIJYgkgyNQIt8g"
    "1R/1HjMerx0NHSwchBu5GoQZZhhKFyEWBxXOE6ASmBFvECwP7Q19DPQKcgnaB4EGeAWPBCsETwSQ"
    "BK4E2gTRBFUElwN4Au8AM/9b/WH7avmj9+H1IfSo8kDxA/DL7pDtXuxp667qYepP6oHqDuvv6xbt"
    "le5P8Azy/PP29RL4SfqO/Pn+hgEPBPMGNQqIDekQTRRdFwkaTRzKHXAefx4RHlgddhxcG2Iaehmf"
    "GNIX/xYmFkIVGxSFEhIRvQ9iDhUNyQuOCnoJTgj8Bn8FzQP2AfD/x/2e+4/5nfch9iL1ivRA9CT0"
    "HvQO9MXzR/NV8grxne8d7k7spepK6fnn0+Yl5rXlUeX05JDkKOTR44njkOPs45HkjuXq5rDo8ep/"
    "7SXw2/L59ff4vfub/osBFgSsBmcJEQzRDswR0BS9F4Ma0BxbHh8fVh8WHwAeghwGGz8ZhRcxFrIU"
    "EBPMEa4QZQ8BDrAMbgtAChMJ7gfTBpUFIwSOAtcAB/8w/WL7xPl4+FL3kfZf9nv22/aF9x34e/jT"
    "+Lr4Mfh592L2zPQ988vxd/A47yPucu0T7bDsM+zN61Drw+oH6j3p3OjK6LLo++jZ6QDrXewp7jTw"
    "P/Jp9LD24/gI+zP9MP8+AXIDsAVCCCsLHQ4zESQUwBbtGEcakBpaGsoZpBhIF9oVaxQcE+gR1BD0"
    "DxYPQQ5bDXMMkgusCrMJxAjRB9MGswVZBOACTgHD/zz+tfxG+wL62fj+98P3tfed98X30fey96X3"
    "TPeB9lz1EPTK8oTxKPDu7uTt7uwU7FPrd+qD6XXoW+c/5nvlM+U15XrlVOaj5zzpDesq7W/v2fEX"
    "9D32Y/iS+sn8FP9nAeIDrQaECXkMnw/LErMVKhgrGnEbxhuhGyQbFhrLGH4XKBbjFMMTjBJbESsQ"
    "5g6QDSIMlwotCX8H8QW4BGQD8gG/AKb/af4n/ez7vfp8+T/4MPc89nT1IfXt9N30BPXz9Lr0tPRi"
    "9IPzk/KE8VrwDe/Q7cDs7esl65LqLere6YbpKenn6NPo++hq6RHq2OoC7IztU+8X8evyzvTY9s74"
    "3fr5/Pr+/gA/A4QFDQjKCkINmw8fEmoUQxaxF6wYEBkCGZgY4xf7FjMWeRWPFKgTEBNjEnkRrhDE"
    "D7kOvA3QDBMMXwuJCgUKnAnxCGQI5gf9Bi0GbQVABDEDfAKlAdQArQC/ALoAxgAVAUABEQGfAPz/"
    "Hf8L/s38Wvvv+Zb4U/dV9o317fRH9Izz9PJL8ofxAvGk8DXwZvAE8ZHxhPLj8yP1m/ZU+Oz5iPso"
    "/cf+bwD6AYIDNwXpBo0ITwr+C44N9w4uENgQ7BC6ECIQ9A6lDWQMzgooCdwHjwY1BekDkwJ+AW8A"
    "Kf8t/mb9i/y/+xL7RPqJ+aD4kfeN9mz1O/RW85/yBvK+8d3xP/LU8qfzk/RY9en1V/Z59jj2z/Uv"
    "9Xn05/NZ88byjfK+8vDyRPOj8+zzI/R49ID0L/QX9ED0gvQ29UL2dfe++Dn6zfuo/Yv/TwEIA6ME"
    "EAaKBx8JhQrVC2cNHg/0ELQSVhS/Fc0WThdYF/EW6RW1FHAT4xE/EPAOwQ2GDF4LTApOCUUISAdw"
    "Bp8F3wQ2BIkD2wIhAlcBegCs/27+KP0K/Ab78fkP+Xn4Nfj+9+D3/vc/+Ev4EfjP92b3mPa59ez0"
    "1/O98u7xBPEB8ELvru4g7nztsewk7K/rJevj6ifriutS7Jvt7+5s8DnyCfTQ9Yb3Rfns+pv8MP7G"
    "/4QBXwMlBR4HUAl/C5oNkQ8vEYESZhOeE4UTDxMfEikRMBASDzcOfQ2EDLwLNwuACpYJtwjUB+4G"
    "AwYZBWMEtAMCA14CtQEgAZwA7f8f/2j+wf37/Hb8I/zk+9L72Pvf+9v7zvui+1v72vo4+nf5i/ip"
    "99X28PUC9TH0ivPZ8iDyefHm8Evw0++V72vvb+/M73zwkvHD8v/zffUa96D4Uvoa/L39iP+MAX8D"
    "cwVZB1wJqgvXDa0PWRHWEgAU4xR4FZIVRRXbFDMUNBMyEkwRShA3Dw8O0QyHCzgKAgnoB5QGRgVP"
    "BG0DaAJ+AaAAt/+t/p39jvxH+9/5qfiW96T2GPa09XL1jvWj9bv13PXp9eb14/V09er0RPSE8+vy"
    "XfLG8XDxT/E28STxMvFq8Z3xjvG58QryYvL78sbzlvS29fX2Mvi1+Tv7mvwW/rP/MAGNAgEEjgUc"
    "B5kISAq9Cx4Nqw4LEAYR3hFGEkoSQxLQERkRcBBzDzsOQA1HDCELEgo1CSoIBAcqBocFzQQiBKED"
    "FwOdAicCnwEMAS8AEv9N/pH9sfwD/JX7Vvte+7b7K/yK/Nv8P/1c/RP9yvx3/M77//od+kb5kfjM"
    "9xf3j/Ye9qj1PPWw9E304vNr8zvzSPNV87XzY/RF9Vj2kvfo+Cj6X/u4/AT+Dv8hAFABgQLQAwsF"
    "PAbRB3IJwgozDGQNFw6VDt0OzA52Dq4NkQxwCzYKvAiOB4AGSwUmBFYDdgJpAZMA9f81/4z+K/6t"
    "/Qj9ZPzv+4v79fpL+tH5Yvn2+MH4svjS+Fz58/lm+v/6nfsW/I382/zz/An9Kv0f/fr89/zb/Lz8"
    "yvy0/ET8APzG+077aWQzICIAAABJRDMDAAAAAAAYVFhYWAAAAA4AAABUWFhYAGlzb21tcDQy"
)



import operator
import struct
from dataclasses import dataclass
from enum import Enum
from functools import reduce


from .ble import BLECommandWord

class OneFingerGesture(Enum):
    NONE = 0x00
    SINGLE_TAP = 0x01
    TAP_AND_HOLD = 0x02
    SWIPE_X_NEG = 0x04
    SWIPE_X_POS = 0x08
    SWIPE_Y_NEG = 0x20
    SWIPE_Y_POS = 0x10

class TwoFingerGesture(Enum):
    NONE = 0x00
    TWO_FINGER_TAP = 0x01
    SCROLL = 0x02
    ZOOM = 0x04

@dataclass
class TouchConfig:
    FORMAT_CFG1 = ">BBHHHHBHBHHH"
    FORMAT_CFG2 = ">BBHHHBBHHHHH"
    FORMAT_CMD  = ">BBHHHHHHHHH"

    swipe_initial_time: int
    swipe_initial_distance: int
    swipe_consecutive_time: int
    swipe_consecutive_distance: int
    swipe_angle: int
    scroll_initial_distance: int
    scroll_angle: int
    zoom_initial_distance: int
    zoom_consecutive_distance: int
    tap_time: int
    tap_distance: int
    hold_time: int
    one_finger_gesture: list[OneFingerGesture]
    two_finger_gesture: list[TwoFingerGesture]

    @classmethod
    def FromJSON(cls, json):
        return cls(
            json["swipe_initial_time"],
            json["swipe_initial_distance"],
            json["swipe_consecutive_time"],
            json["swipe_consecutive_distance"],
            json["swipe_angle"],
            json["scroll_initial_distance"],
            json["scroll_angle"],
            json["zoom_initial_distance"],
            json["zoom_consecutive_distance"],
            json["tap_time"],
            json["tap_distance"],
            json["hold_time"],
            [OneFingerGesture(f) for f in json["one_finger_gesture"]],
            [TwoFingerGesture(f) for f in json["two_finger_gesture"]],
        )
    
    @classmethod
    def Default(cls):
        return cls(
            swipe_initial_time=200,
            swipe_initial_distance=25,
            swipe_consecutive_time=150,
            swipe_consecutive_distance=25,
            swipe_angle=23,
            scroll_initial_distance=50,
            scroll_angle=37,
            zoom_initial_distance=50,
            zoom_consecutive_distance=5,
            tap_time=300,
            tap_distance=75,
            hold_time=10,
            one_finger_gesture=[OneFingerGesture.SINGLE_TAP, OneFingerGesture.TAP_AND_HOLD],
            two_finger_gesture=[TwoFingerGesture.TWO_FINGER_TAP]
        )
    
    @classmethod
    def High(cls):
        c = cls.Default()
        c.tap_time = 150
        c.tap_distance = 50
        return c

    @classmethod
    def Medium(cls):
        return cls.Default()

    @classmethod
    def Low(cls):
        c = cls.Default()
        c.tap_time = 600
        c.tap_distance = 150
        c.hold_time = 50
        return c

    def toJSON(self):
        return {
            "swipe_initial_time": self.swipe_initial_time,
            "swipe_initial_distance": self.swipe_initial_distance,
            "swipe_consecutive_time": self.swipe_consecutive_time,
            "swipe_consecutive_distance": self.swipe_consecutive_distance,
            "swipe_angle": self.swipe_angle,
            "scroll_initial_distance": self.scroll_initial_distance,
            "scroll_angle": self.scroll_angle,
            "zoom_initial_distance": self.zoom_initial_distance,
            "zoom_consecutive_distance": self.zoom_consecutive_distance,
            "tap_time": self.tap_time,
            "tap_distance": self.tap_distance,
            "hold_time": self.hold_time,
            "one_finger_gesture": [ofg.value for ofg in self.one_finger_gesture],
            "two_finger_gesture": [tfg.value for tfg in self.two_finger_gesture],
        }

    def set_sensitivity(self, sensitivity: int):
        if sensitivity == 1:
            self.tap_time = 600
            self.tap_distance = 150
            self.hold_time = 50
        elif sensitivity == 2:
            self.tap_time = 300
            self.tap_distance = 75
        else:
            self.tap_time = 150
            self.tap_distance = 50

    def toBytes(self) -> tuple[bytes, bytes, bytes]:
        cfg1 = struct.pack(self.FORMAT_CFG1,
            BLECommandWord.UPDATE_TOUCH1.value,
            16,
            self.swipe_initial_time,
            self.swipe_initial_distance,
            self.swipe_consecutive_time,
            self.swipe_consecutive_distance,
            self.swipe_angle,
            self.scroll_initial_distance,
            self.scroll_angle,
            self.zoom_initial_distance,
            self.zoom_consecutive_distance,
            0
        )

        cfg2 = struct.pack(self.FORMAT_CFG2,
            BLECommandWord.UPDATE_TOUCH2.value,
            8,
            self.tap_time,
            self.tap_distance,
            self.hold_time,
            reduce(operator.ior, [g.value for g in self.one_finger_gesture]) if self.one_finger_gesture else 0,
            reduce(operator.ior, [g.value for g in self.two_finger_gesture]) if self.two_finger_gesture else 0,
            0,
            0,
            0,
            0,
            0,
        )

        save = struct.pack(self.FORMAT_CMD, BLECommandWord.WRITE_TOUCH.value,1,0,0,0,0,0,0,0,0,0)
        return cfg1, cfg2, save

    def update(self, cfg1: bytes, cfg2: bytes):
        word_cfg1, length_cfg1, *data_cfg1 = struct.unpack(self.FORMAT_CFG1, cfg1)
        word_cfg2, length_cfg2, *data_cfg2 = struct.unpack(self.FORMAT_CFG2, cfg2)

        if BLECommandWord(word_cfg1) != BLECommandWord.UPDATE_TOUCH1 or \
            BLECommandWord(word_cfg2) != BLECommandWord.UPDATE_TOUCH2 or \
            length_cfg1 != 16 or \
            length_cfg2 != 8:
            return False

        self.swipe_initial_time = data_cfg1[0]
        self.swipe_initial_distance = data_cfg1[1]
        self.swipe_consecutive_time = data_cfg1[2]
        self.swipe_consecutive_distance = data_cfg1[3]
        self.swipe_angle = data_cfg1[4]
        self.scroll_initial_distance = data_cfg1[5]
        self.scroll_angle = data_cfg1[6]
        self.zoom_initial_distance = data_cfg1[7]
        self.zoom_consecutive_distance = data_cfg1[8]
        self.tap_time = data_cfg2[0]
        self.tap_distance = data_cfg2[1]
        self.hold_time = data_cfg2[2]
        self.one_finger_gesture = [g for g in OneFingerGesture if bool(g.value & data_cfg2[3])]
        self.two_finger_gesture = [g for g in TwoFingerGesture if bool(g.value & data_cfg2[4])]

        return True


from dataclasses import dataclass
from enum import Enum

from .ble import BLESelector
from .gesture import GestureConfig, Gesture
from .touch import TouchConfig, OneFingerGesture, TwoFingerGesture

class Hand(Enum):
    RIGHT = "right"
    LEFT = "left"

@dataclass
class Acceleration:
    x: float
    y: float
    z: float

    def toJSON(self) -> dict:
        return {
            "x": self.x,
            "y": self.y,
            "z": self.z,
        }

@dataclass
class Angle:
    roll: float
    pitch: float
    yaw: float

    def toJSON(self) -> dict:
        return {
            "roll": self.roll,
            "pitch": self.pitch,
            "yaw": self.yaw,
        }

@dataclass
class Gyro:
    x: float
    y: float
    z: float

    def toJSON(self) -> dict:
        return {
            "x": self.x,
            "y": self.y,
            "z": self.z,
        }

@dataclass
class Touch:
    one_finger: OneFingerGesture
    two_finger: TwoFingerGesture
    x_pos: float
    y_pos: float

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, Touch):
            return (
                self.one_finger == __value.one_finger 
                and self.two_finger == __value.two_finger
                and self.x_pos == __value.x_pos
                and self.y_pos == __value.y_pos
            )
        
        return False
    
    def toJSON(self) -> dict:
        return {
            "one_finger": self.one_finger.value,
            "two_finger": self.two_finger.value,
            "x_pos": self.x_pos,
            "y_pos": self.y_pos
        }
    
@dataclass
class TSkinState:
    connected: bool
    sleep: bool
    battery: float | None
    selector: BLESelector | None
    touch: Touch | None
    angle: Angle | None
    gesture: Gesture | None

    def toJSON(self) -> dict:
        return {
            "connected": self.connected,
            "sleep": self.sleep,
            "battery": self.battery,
            "selector": self.selector.value if self.selector else None,
            "touch": self.touch.toJSON() if self.touch else None,
            "angle": self.angle.toJSON() if self.angle else None,
            "gesture": self.gesture.toJSON() if self.gesture else None,
        }
    

@dataclass
class TSkinConfig:
    address: str
    hand: Hand
    name: str = "Tactigon"
    touch_config: TouchConfig | None = None
    gesture_config: GestureConfig | None = None

    @classmethod
    def FromJSON(cls, json: dict):
        return cls(
            json["address"], 
            Hand(json["hand"]),
            json["name"] if "name" in json else cls.name,
            TouchConfig.FromJSON(json["touch_config"]) if "touch_config" in json and json["touch_config"] is not None else None,
            GestureConfig.FromJSON(json["gesture_config"]) if "gesture_config" in json and json["gesture_config"] is not None else None,
        )
    
    def toJSON(self) -> dict:
        return {
            "address": self.address,
            "hand": self.hand.value,
            "name": self.name,
            "touch_config": self.touch_config.toJSON() if self.touch_config else None,
            "gesture_config": self.gesture_config.toJSON() if self.gesture_config else None,
        }


import time
from src.tactigon_gear.tskin import TSkin
from src.tactigon_gear.models.tskin import TSkinConfig, Hand, Acceleration, Gyro, Angle
from src.tactigon_gear.models.gesture import GestureConfig
from src.tactigon_gear.models.touch import OneFingerGesture
from src.tactigon_gear.hal.ble import Ble

# C0:83:4B:32:4E:36
# BE5BDEB7-72CC-4221-A44D-103795E74335

def test():
    # model_folder = path.join(path.abspath("."), "data", "models", "MODEL_01_R")

    gmodel = None

    with TSkin(TSkinConfig("C0:83:4F:34:28:38", Hand.RIGHT, "TEST", None, gmodel), True) as tskin:
        print("connecting tskin", tskin)

        while not tskin.connected:
            time.sleep(1)

        while True:

            if not tskin.connected:
                print("Disconnected...")
                time.sleep(2)
                continue

            if tskin.sleep:
                print("Sleeping...")
                time.sleep(2)
                continue

            if not tskin.calibrated:
                print("Calibrating... Move around")
                time.sleep(1)
                continue

            acc = tskin.acceleration
            gyro = tskin.gyro
            angle = tskin.angle

            if acc and gyro and angle:
                print(f"{round(angle.roll, 2)}\t{round(angle.pitch, 2)}\t{round(angle.yaw, 2)}\t{round(acc.x, 3)}\t{round(acc.y, 3)}\t{round(acc.z, 3)}\t{round(gyro.x, 3)}\t{round(gyro.y, 3)}\t{round(gyro.z, 3)}")

            time.sleep(tskin.TICK)

    print("exit")


if __name__ == "__main__":
    test()


import setup_path
import time
import logging
from datetime import datetime

from tactigon_gear.tskin import TSkin
from tactigon_gear.models.tskin import Hand, TSkinConfig
from tactigon_gear.models.touch import OneFingerGesture, TwoFingerGesture
from tactigon_gear.models.ble import BLESelector

# C0:83:3F:39:21:57
# 296A6374-C160-2914-7A19-A8B8861DC309

logging.basicConfig(
    filename="log.log", 
    level=logging.DEBUG
)
# logging.getLogger("bleak").setLevel(logging.INFO)



def test():
    last_update = datetime.now()
    audio_sec = 5
    tskin_cfg = TSkinConfig(
        "C0:83:3F:39:21:57",
        Hand.RIGHT
    )

    with TSkin(tskin_cfg, debug=True) as tskin:
        i = 0

        while True:
            if not tskin.connected:
                print("Connecting..")
                while not tskin.connected:
                    time.sleep(tskin.TICK)

                print("Connected!")

            if i > 5:
                break

            state = tskin.state

            if state.selector == BLESelector.AUDIO:
                if (datetime.now()-last_update).total_seconds() > audio_sec:
                    tskin.select_sensors()

            if state.sleep:
                print("Device in sleep...")
                lll = 0
                while tskin.sleep:
                    lll += 1
                    time.sleep(tskin.TICK)

                print(lll * tskin.TICK)
                continue

            if state.touch:
                if state.touch.one_finger == OneFingerGesture.SINGLE_TAP:
                    last_update = datetime.now()
                    tskin.select_audio()
                elif state.touch.one_finger == OneFingerGesture.TAP_AND_HOLD:
                    i += 1

            print(tskin.acceleration, tskin.battery)

            time.sleep(tskin.TICK)

if __name__ == "__main__":
    test()


import time
import wave
import logging
# import msvcrt

from os import path
from multiprocessing import Process, Pipe, Event, log_to_stderr, Queue
from multiprocessing.synchronize import Event as EventClass
from multiprocessing.connection import _ConnectionBase

from datetime import datetime
from typing import Optional, Generator

from src.tactigon_gear.tskin import TSkin
from src.tactigon_gear.models.tskin import TSkinConfig, Hand
from src.tactigon_gear.models.ble import TBleSelector
from src.tactigon_gear.models.gesture import GestureConfig
from src.tactigon_gear.models.touch import TouchConfig, OneFingerGesture

# from src.tactigon_gear import TSkin, TSkinConfig, Hand, OneFingerGesture, TwoFingerGesture
# from src.tactigon_gear.models import TBleSelector, GestureConfig, TouchConfig

class Audio(Process):
    _TICK: float = 0.02
    _PIPE_TIMEOUT: float = 0.1
    _SAMPLE_RATE: float = 16000

    def __init__(self, stop: EventClass, pipe: _ConnectionBase, listen: EventClass, debug: bool = False):
        Process.__init__(self)
        self.logger = log_to_stderr()
        if debug:
            self.logger.setLevel(logging.DEBUG)

        self.stop = stop
        self.pipe = pipe
        self.listen = listen

    def vad_collector(self) -> Generator[bytes, bytes, None]:
        seconds = 5
        frame_length = 80
        num_samples = 2 * self._SAMPLE_RATE * seconds / frame_length

        frame = b''
        num_frames = 0
    
        while True:
            if num_frames >= num_samples:
                break
            
            if not self.pipe.poll():
                time.sleep(self._TICK)
                continue
            
            frame = self.pipe.recv_bytes()
            num_frames += 1

            yield frame

        return None

    def run(self):
        self.logger.debug("[Audio] Process started!")
        while not self.stop.is_set():
            if not self.listen.is_set():
                time.sleep(self._TICK)
                continue
            self.logger.debug("[Audio] Listening started!")
            i = 0
            with wave.open("test_audio.wav", "wb") as test_file:
                test_file.setsampwidth(2)
                test_file.setnchannels(1)
                test_file.setframerate(16000)

                for frame in self.vad_collector():
                    if frame is None:
                        break
                    test_file.writeframes(frame)
                    i += 1

            self.listen.clear()
            self.logger.debug("[Audio] Listening ended! Processes %i packets", i)

class TSkin_Audio(TSkin):
    def __init__(self, config: TSkinConfig, debug: bool = False):
        TSkin.__init__(self, config, debug)
        self._audio_rx, self._audio_tx = Pipe(duplex=False)
        self._audio_stop = Event()
        self._listen_event = Event()
        self.audio = Audio(self._audio_stop, self._audio_rx, self._listen_event, debug)

    def start(self):
        self.audio.start()
        TSkin.start(self)

    def join(self, timeout: Optional[float] = None):
        self.select_sensors()
        if self._listen_event.is_set():
            self.stop_listen()

        self._audio_stop.set()
        self.audio.join(timeout)
        self.audio.terminate()
        TSkin.join(self, timeout)

    def listen(self):
        logging.debug("[TSkin] Starting listen..")
        self._listen_event.set()
        self.select_audio()

        while self._listen_event.is_set():
            time.sleep(self.TICK)

        logging.debug("[TSkin] Stopped listen..")
        self.select_sensors()

    def stop_listen(self):
        logging.debug("[TSkin] Stopping listen..")
        self.select_sensors()
        self._listen_event.clear()
        self.clear_audio_pipe()

    def clear_audio_pipe(self):
        i = 0
        while self._audio_rx.poll():
            _ = self._audio_rx.recv_bytes()
            i += 1

        logging.debug("[Audio] Cleare %i packet from pipe", i)

def test():
    model_folder = path.join(path.abspath("."), "data", "models", "MODEL_01_R")

    gconfig = GestureConfig(
        path.join(model_folder, "model.pickle"), 
        path.join(model_folder, "encoder.pickle"),
        "test",
        datetime.now()
    )

    tconfig = TouchConfig.Default()

    tskin_cfg = TSkinConfig(
        #"C0:83:1F:34:23:38", #"E6:00:EC:EA:A8:A1",
        "C0:83:3F:39:21:57",
        Hand.RIGHT,
        touch_config=tconfig,
        gesture_config=gconfig
    )

    with TSkin_Audio(tskin_cfg, False) as tskin:
        print("connecting")
        while not tskin.connected:
            time.sleep(0.5)
        print("connected")
        
        while True:
            # if msvcrt.kbhit():
            #     key = msvcrt.getwch()

            #     if key == "q":
            #         break
            #     elif key == "l":
            #         tskin.listen()
            #     elif key == "c":
            #         tskin.calibrate()

            if not tskin.connected:
                print("Disconnected...", tskin.sleep)
                time.sleep(2)
                continue

            if tskin.sleep:
                print("Sleeping...")
                time.sleep(2)
                continue

            if not tskin.calibrated:
                print("Calibrating... Move around")
                time.sleep(1)
                continue

            if tskin.selector == TBleSelector.AUDIO:
                print("Listening...")
                time.sleep(0.1)
                continue


            acc = tskin.acceleration
            gyro = tskin.gyro
            angle = tskin.angle
            batt = tskin.battery
            gest = tskin.gesture
            t = tskin.touch

            if t:
                if t.one_finger == OneFingerGesture.SINGLE_TAP:
                    print("Listening")
                    tskin.listen()
                    print("Done!")
                else:
                    print("Exiting...")
                    break

            if gest:
                print(gest)

            # if acc and gyro and angle:
            #     print(f"{round(angle.roll, 2)}\t{round(angle.pitch, 2)}\t{round(angle.yaw, 2)}\t{round(acc.x, 3)}\t{round(acc.y, 3)}\t{round(acc.z, 3)}\t{round(gyro.x, 3)}\t{round(gyro.y, 3)}\t{round(gyro.z, 3)}\t{batt}\t{gest.gesture if gest else ''}\t{touch}")                
            #     pass
            
            time.sleep(0.02)


if __name__ == "__main__":
    test()


import time
import csv
from src.tactigon_gear.tskin import TSkin
from src.tactigon_gear.models.tskin import Hand, TSkinConfig
from src.tactigon_gear.models.touch import OneFingerGesture, TwoFingerGesture

# C0:83:4B:32:4E:36
# BE5BDEB7-72CC-4221-A44D-103795E74335

def test():
    tskin_cfg = TSkinConfig(
        "C0:83:22:30:55:58",
        Hand.RIGHT
    )

    tskin = TSkin(tskin_cfg, debug=True)
    tskin.start()

    i = 0

    with open("test_ble.csv", "w", newline='', encoding='utf-8') as test_file:
        test_writer = csv.writer(test_file)
        test_writer.writerow(["battery", "roll", "pitch", "yaw", "touch"])
    

        while True:
            if not tskin.connected:
                print("Connecting..")
                time.sleep(0.5)
                continue

            if i > 5:
                break

            batt = tskin.battery
            a = tskin.angle
            t = tskin.touch
            acc = tskin.acceleration
            gyro = tskin.gyro

            result = [str(batt)]
            result.extend([str(a.roll), str(a.pitch), str(a.yaw)] if a else ["","",""])
            result.extend([t.one_finger.name, t.two_finger.name, str(t.x_pos), str(t.y_pos)] if t else ["","","", ""])

            test_writer.writerow(result)

            print(batt, a, t)

            if t:
                if t.one_finger == OneFingerGesture.TAP_AND_HOLD:
                    i += 1

            time.sleep(0.01)

    tskin.terminate()

        


if __name__ == "__main__":
    test()


__all__ = ['ADPCMEngine', 'PacketManager', 'TactigonGesture', 'TactigonRecorder']

from ..middleware.adpcm_engine import ADPCMEngine
from ..middleware.packet_manager import PacketManager
from ..middleware.tactigon_gesture import TactigonGesture
from ..middleware.tactigon_recorder import TactigonRecorder


import selectors
import logging
import socket
from threading import Thread, Event

from typing import Any, Optional, Dict, Tuple

from ..models.socket import SocketCommand
from ..middleware.packet_manager import PacketManager

class Listener(Thread):
    _listener: socket.socket
    _stop_event: Event

    _connected: bool = False

    def __init__(self, url: str, port: int):
        Thread.__init__(self)
        self._stop_event = Event()
        self._address = (url, port)
        self._logger = logging.getLogger(Listener.__name__)
        self._packet_manager = PacketManager(self._logger)

        self._selector = selectors.DefaultSelector()
        self._clients: Dict[socket.socket, Tuple[str, int]] = {}

    def run(self):
        self._logger.debug("Started!")
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as self._listener:
            self._listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._listener.bind(self._address)
            self._listener.listen()
            self._listener.setblocking(False)
            self._selector.register(self._listener, selectors.EVENT_READ)
            self._logger.debug("Listener created on %s", self._address)

            try:
                while not self._stop_event.is_set():
                    events = self._selector.select(timeout=0.1)
                    for key, mask in events:
                        sock: socket.socket = key.fileobj #type: ignore
                        if sock is self._listener:
                            conn, addr = self._listener.accept()
                            conn.setblocking(False)
                            self._selector.register(conn, selectors.EVENT_READ)
                            self._clients[conn] = addr
                            self._logger.debug("Connected client %s", addr)
                        else:
                            try:
                                packet_size_bytes = self._packet_manager.recv_bytes(sock, 4)
                                if not packet_size_bytes:
                                    raise ConnectionError("Client disconnected")
                                bufsize = int.from_bytes(packet_size_bytes, "big")
                                packet_body = self._packet_manager.recv_bytes(sock, bufsize)
                                command, payload = self._packet_manager.unpack_command(packet_body)
                                self._logger.debug("Got command %s data %s", command, payload)
                                self.task(command, payload)
                            except ConnectionError:
                                addr = self._clients.pop(sock, ("Unknown", 0))
                                self._logger.debug("Client %s disconnected", addr)
                                self.task(SocketCommand.DISCONNECTED, {})
                                self._selector.unregister(sock)
                                sock.close()
                            except Exception as e:
                                self._logger.error(e)

            except Exception as e:
                self._logger.error(e)
            finally:
                for conn in list(self._clients.keys()):
                    try:
                        self._selector.unregister(conn)
                        conn.close()
                    except Exception:
                        pass
                try:
                    self._selector.unregister(self._listener)
                except Exception:
                    pass
                self._logger.debug("Stopped!")
        
    def task(self, command: SocketCommand, payload: dict):
        pass

    def send(self, command: SocketCommand, payload: dict | list[bytes]):
        buff = self._packet_manager.pack_command(command, payload)
        for client in self._clients:
            try:
                client.sendall(buff)
            except Exception as e:
                self._clients.pop(client)
                self._logger.error("Cannot send data. %s", e)        

    def join(self, timeout: Optional[float] = None):
        self._logger.debug("Called stop...")
        self._stop_event.set()
        Thread.join(self, timeout)
        self._logger.debug("Joined!")





import logging
import struct
import math
import asyncio
import time

from bleak import BleakClient
from threading import Thread, Event, Lock
from queue import Queue
from multiprocessing.connection import _ConnectionBase

from typing import Optional, Tuple

from ..models.ble import BLECommand, BLECommandMask, BLECommandWord, BLESelector, BLECommands
from ..models.tskin import Hand, Angle, Acceleration, Touch, Gyro
from ..models.touch import TouchConfig, OneFingerGesture, TwoFingerGesture

from ..utils import AdvancedQueue
from ..middleware.adpcm_engine import ADPCMEngine

class Ble(Thread):
    TICK: float = 0.02
    _RECONNECT_TIMEOUT: float = 0.1
    STATUS_UUID: str =          "bea5760d-503d-4920-b000-101e7306b012"
    SENSORS_UUID: str =         "bea5760d-503d-4920-b000-101e7306b005"
    TOUCHPAD_UUID: str =        "bea5760d-503d-4920-b000-101e7306b009"
    TOUCH_CONFIG_UUID: str =    "bea5760d-503d-4920-b000-101e7306b011"

    AUDIO_DATA_UUID: str =      "08000000-0001-11e1-ac36-0002a5d5c51b"
    AUDIO_SYNC_UUID: str =      "40000000-0001-11e1-ac36-0002a5d5c51b"

    TERMINAL_UUID: str =        "00000001-000e-11e1-ac36-0002a5d5c51b"
    CONFIG_UUID: str =          "00000002-000f-11e1-ac36-0002a5d5c51b"

    address: str
    hand: Hand
    touch_config: TouchConfig
    debug: bool

    _stop_event: Event
    _client_connected: bool = False
    selector: BLESelector
    _sensor_tx: Optional[_ConnectionBase] = None
    _angle_tx: Optional[_ConnectionBase] = None
    _audio_tx: Optional[_ConnectionBase] = None
    _update: Lock
    _update_touch: Lock
    _update_selector: Lock
    _sleep: Event
    _charging: Event
    _calibrated_sensorfusion: Event
    _calibrated_compass: Event

    _fw_version: Optional[str] = None
    _angle: Optional[Angle] = None
    _acceleration: Optional[Acceleration] = None
    _gyro: Optional[Gyro] = None
    _battery: float = 0
    _touch: Optional[Touch] = None

    _writes: AdvancedQueue
    _reads: Queue
    _ble_logger: logging.Logger

    adpcm_engine: ADPCMEngine

    def __init__(self, address: str, hand: Hand, touch_config: Optional[TouchConfig] = None, debug: bool = False):
        Thread.__init__(self)
        self._ble_logger = logging.getLogger(Ble.__name__)
        self.address = address
        self.hand = hand
        
        if touch_config:
            self.touch_config = touch_config
        else:
            self.touch_config = TouchConfig.Default()

        self.debug = debug

        self._sleep = Event()
        self._charging = Event()
        self._calibrated_sensorfusion = Event()
        self._calibrated_compass = Event()
        self._stop_event = Event()
        self._update = Lock()
        self._update_touch = Lock()
        self._update_selector = Lock()
        self.selector = BLESelector.SENSORS
        self.adpcm_engine = ADPCMEngine()

        self._writes = AdvancedQueue()
        self._reads = Queue()

        self.update_touch(self.touch_config)
        self._calibrated_sensorfusion.set()
        self._calibrated_compass.set()

    @staticmethod
    def gravity_comp(hand: Hand, accX: float, accY: float, accZ: float, gyroX: float, gyroY: float, gyroZ: float, roll: float, pitch: float, yaw: float):
        """gravity compensation"""
        G_CONST = 9.81
        ANG_TO_RAD = math.pi / 180
        ACC_RATIO = 1000
        VEL_RATIO = 30

        if hand == Hand.LEFT:
            accX = -accX / ACC_RATIO
            accY = -accY / ACC_RATIO
            accZ = -accZ / ACC_RATIO

            gyroX = -gyroX / VEL_RATIO
            gyroY = -gyroY / VEL_RATIO
            gyroZ = -gyroZ / VEL_RATIO

            _pitch = roll * ANG_TO_RAD
            _roll = pitch * ANG_TO_RAD

        else:
            accX = accX / ACC_RATIO
            accY = accY / ACC_RATIO
            accZ = -accZ / ACC_RATIO

            gyroX = gyroX / VEL_RATIO
            gyroY = gyroY / VEL_RATIO
            gyroZ = -gyroZ / VEL_RATIO

            _pitch = -roll * ANG_TO_RAD
            _roll = -pitch * ANG_TO_RAD

        if accZ == 0:
            beta = math.pi / 2
        else:
            beta = math.atan(
                math.sqrt(math.pow(accX, 2) + math.pow(accY, 2)) / accZ
            )

        accX = accX - G_CONST * math.sin(_roll)
        accY = accY + G_CONST * math.sin(_pitch)
        accZ = accZ - G_CONST * math.cos(beta)

        return accX, accY, accZ, gyroX, gyroY, gyroZ, roll, pitch, yaw

    @property
    def firmware_version(self) -> Optional[str]:
        return self._fw_version

    @property
    def connected(self) -> bool:
        return self._client_connected
    
    @property
    def angle(self) -> Optional[Angle]:
        if self.sleep:
            return None
        
        with self._update:
            angle = self._angle
        return angle
    
    @property
    def acceleration(self) -> Optional[Acceleration]:
        if self.sleep:
            return None
        
        with self._update:
            acc = self._acceleration
        return acc

    @property
    def gyro(self) -> Optional[Gyro]:
        if self.sleep:
            return None
        
        with self._update:
            gyro = self._gyro
        return gyro
    
    @property
    def touch(self) -> Optional[Touch]:
        if self.sleep:
            return None
        
        with self._update_touch:
            touch = self._touch
            self._touch = None
        return touch
    
    @property
    def sleep(self) -> bool:
        return self._sleep.is_set()
        
    @property
    def battery(self) -> Optional[float]:
        if self.sleep:
            return None
        
        return self._battery

    @property
    def calibrated(self) -> bool:
        return self._calibrated_compass.is_set() and self._calibrated_sensorfusion.is_set()

    def on_terminal(self, char, data: bytearray):
        print("TERMINAL", data)

    def on_config(self, char, data: bytearray):
        _tick = int.from_bytes(data[0:2], 'little')
        command = BLECommandWord(data[6])
        payload = int.from_bytes(data[7:], 'little')
        _mask = int.from_bytes(data[2:6], 'big')
        mask = BLECommandMask(bytes([_mask >> 24 & 0xFF, _mask >> 16 & 0xFF, _mask >> 8 & 0xFF, _mask & 0xFF]))

        if mask == BLECommandMask.SENSORFUSION:
            if command == BLECommandWord.STATUS:
                if payload == 100:
                    self._calibrated_sensorfusion.set()
                else:
                    self._calibrated_sensorfusion.clear()
        elif mask == BLECommandMask.ECOMPASS:
            if command == BLECommandWord.STATUS:
                if payload == 100:
                    self._calibrated_compass.set()
                else:
                    self._calibrated_compass.clear()

        self._ble_logger.debug("Config %i %s %s %s", _tick, mask, command, payload)

    def handle_status(self, char, data: bytearray):
        fw_major, fw_minor, fw_patch, fw_dev, sleep_armed = struct.unpack("cccc?", data[0:5])
        self._fw_version = f"{fw_major.decode()}.{fw_minor.decode()}.{fw_patch.decode()}.{fw_dev.decode()}"
        
        if sleep_armed:
            self._sleep.set()
        else:
            self._sleep.clear()

        self._ble_logger.debug("FW version: %s, sleep %s", self._fw_version, self.sleep)
        
    def handle_audio_sync(self, char, data: bytearray):
        pass

    def handle_audio(self, char, data: bytearray):
        if self._audio_tx:
            self._audio_tx.send_bytes(self.adpcm_engine.extract_data(data))

    def handle_sensors(self, char, data:bytearray):       
        accX, accY, accZ, gyroX, gyroY, gyroZ, roll, pitch, yaw, battery = struct.unpack("hhhhhhhhhh", data)
        accX, accY, accZ, gyroX, gyroY, gyroZ, roll, pitch, yaw = self.gravity_comp(self.hand, accX, accY, accZ, gyroX, gyroY, gyroZ, roll, pitch, yaw)

        with self._update:
            self._angle = Angle(roll, pitch, yaw)
            self._acceleration = Acceleration(accX, accY, accZ)
            self._gyro = Gyro(gyroX, gyroY, gyroZ)
            self._battery = battery/1000

        if self.debug:
            self._ble_logger.debug("Angle: %f,%f,%f",roll, pitch, yaw)
            self._ble_logger.debug("Acceleration: %f,%f,%f",accX, accY, accZ)
            self._ble_logger.debug("Gyro: %f,%f,%f",gyroX, gyroY, gyroZ)
            self._ble_logger.debug("Battery: %f", battery/1000)

        if self._sensor_tx:
            self._sensor_tx.send([accX, accY, accZ, gyroX, gyroY, gyroZ])

        if self._angle_tx:
            self._angle_tx.send([roll, pitch, yaw])

    def handle_touchpad(self, char, data: bytearray):
        with self._update_touch:
            one_finger = OneFingerGesture(int.from_bytes(data[0:1], "big"))
            two_finger = TwoFingerGesture(int.from_bytes(data[1:2], "big"))
            if one_finger is not OneFingerGesture.NONE or two_finger is not TwoFingerGesture.NONE:
                self._touch = Touch(
                    one_finger,
                    two_finger,
                    float(struct.unpack("h", data[2:4])[0]),
                    float(struct.unpack("h", data[4:6])[0])
                )

    def start(self):
        self._ble_logger.debug("BLE starting on address %s", self.address)
        Thread.start(self)

    def join(self, timeout: Optional[float] = None):
        self._ble_logger.debug("Stopping BLE on address %s", self.address)
        self._stop_event.set()
        Thread.join(self, timeout)

    def run(self):
        asyncio.run(self.task())

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *attr):
        self.join()

    async def task(self):
        running_selector: Optional[BLESelector] = None

        while not self._stop_event.is_set():
            self._client_connected = False
            
            try:
                running_selector = None

                async with BleakClient(self.address, disconnected_callback=self.on_disconnect) as client:
                    self._ble_logger.debug("%s connected", client)

                    await client.start_notify(self.TERMINAL_UUID, self.on_terminal)
                    await client.start_notify(self.CONFIG_UUID, self.on_config)
                    await client.start_notify(self.STATUS_UUID, self.handle_status)
                    await client.start_notify(self.TOUCHPAD_UUID, self.handle_touchpad)

                    self._client_connected = True

                    while self.connected:
                        if self._stop_event.is_set():
                            await client.disconnect()
                            break

                        if not self.sleep:
                            while True:
                                try:
                                    cmd: BLECommand = self._writes.get_nowait()
                                except:
                                    break

                                if cmd.command == BLECommands.WRITE and cmd.payload:
                                    await client.write_gatt_char(cmd.characteristic, cmd.payload)
                                elif cmd.command == BLECommands.READ:
                                    data = await client.read_gatt_char(cmd.characteristic)
                                    if data:
                                        self._reads.put(data)
                            
                        if running_selector != self.selector:
                            with self._update_selector:
                                if running_selector == BLESelector.SENSORS:
                                    await client.stop_notify(self.SENSORS_UUID)

                                    self._update.acquire()
                                    self._angle = None
                                    self._acceleration = None
                                    self._gyro = None
                                    self._update.release()
                                    self._ble_logger.debug("Stopped notification on sensors (%s)", self.SENSORS_UUID)
                                elif running_selector == BLESelector.AUDIO:
                                    await client.stop_notify(self.AUDIO_DATA_UUID)
                                    await client.stop_notify(self.AUDIO_SYNC_UUID)
                                    self._ble_logger.debug("Stopped notification on AUDIO (%s)", self.AUDIO_DATA_UUID)

                                running_selector = self.selector

                                if running_selector == BLESelector.SENSORS:
                                    await client.start_notify(self.SENSORS_UUID, self.handle_sensors)

                                    self._ble_logger.debug("Started notification on sensors (%s)", self.SENSORS_UUID)
                                elif running_selector == BLESelector.AUDIO:
                                    await client.start_notify(self.AUDIO_SYNC_UUID, self.handle_audio_sync) 
                                    await client.start_notify(self.AUDIO_DATA_UUID, self.handle_audio)

                                    self._ble_logger.debug("Started notification on AUDIO (%s)", self.AUDIO_DATA_UUID)

                        await asyncio.sleep(self._RECONNECT_TIMEOUT)
            
            except Exception as e:
                self._ble_logger.error("BLE error: %s", e)
                await asyncio.sleep(self._RECONNECT_TIMEOUT)

    def on_disconnect(self, client: BleakClient):
        self._ble_logger.debug("Client %s disconnected", client)
        self._client_connected = False

    def set_touch(self, config: TouchConfig, timeout: float = 5) -> bool:
        cfg1, cfg2, save = config.toBytes()

        ref1 = self._writes.put(BLECommand(BLECommands.WRITE, self.TOUCH_CONFIG_UUID, cfg1))
        ref2 = self._writes.put(BLECommand(BLECommands.WRITE, self.TOUCH_CONFIG_UUID, cfg2))
        refsave = self._writes.put(BLECommand(BLECommands.WRITE, self.TOUCH_CONFIG_UUID, save))

        _t = 0
        while _t < timeout:
            if self._writes.check(ref1) and self._writes.check(ref2) and self._writes.check(refsave):
                return True
            
            _t += self._RECONNECT_TIMEOUT
            time.sleep(self._RECONNECT_TIMEOUT)
            
        return False

    def get_touch(self, timeout: float = 5) -> Optional[Tuple[bytes, bytes]]:
        self._writes.put(BLECommand(BLECommands.READ, self.TOUCH_CONFIG_UUID))
        self._writes.put(BLECommand(BLECommands.READ, self.TOUCH_CONFIG_UUID))

        cfg1 = None
        cfg2 = None

        _t = 0
        while _t < timeout:
            try:
                resp: bytes = self._reads.get_nowait()
                if cfg1 == None:
                    cfg1 = resp
                else:
                    cfg2 = resp
            except:
                pass

            if cfg1 and cfg2:
                return cfg1, cfg2

            _t += self._RECONNECT_TIMEOUT
            time.sleep(self._RECONNECT_TIMEOUT)
            
        return None
    
    def update_touch(self, touch_config: TouchConfig, timeout: float = 5) -> bool:
        if self.set_touch(touch_config, timeout):
            res = self.get_touch(timeout)
            return self.touch_config.update(*res) if res else False

        return False
    
    def set_touch_sensitivity(self, sensitivity: int = 2, timeout: float = 5):
        self.touch_config.set_sensitivity(sensitivity)
        return self.set_touch(self.touch_config, timeout)

    def calibrate(self, timeout: float = 5) -> bool:
        cmd = BLECommandMask.SENSORFUSION.value + bytes(BLECommandWord.START.value) + b'\x00'
        ret = self._writes.put(BLECommand(BLECommands.WRITE, self.CONFIG_UUID, cmd))

        _t = 0
        while _t < timeout:
            if self._writes.check(ret):
                return True
            
            _t += self._RECONNECT_TIMEOUT
            time.sleep(self._RECONNECT_TIMEOUT)
            
        return False

    def select_sensors(self):
        self._update_selector.acquire()
        self.selector = BLESelector.SENSORS
        self._update_selector.release()

    def select_audio(self):
        self._update_selector.acquire()
        self.selector = BLESelector.AUDIO
        self._update_selector.release()

    def connect(self):
        self.start()

    def disconnect(self):
        self.join()

    def terminate(self):
        self.join()


import time
import datetime
from os import path
from src.tactigon_gear.tskin import TSkin
from src.tactigon_gear.models.tskin import TSkinConfig, Hand
from src.tactigon_gear.models.gesture import GestureConfig
from src.tactigon_gear.models.touch import OneFingerGesture

# C0:83:4B:32:4E:36
# BE5BDEB7-72CC-4221-A44D-103795E74335

def test():
    model_folder = path.join(path.abspath("."), "data", "models", "MODEL_01_R")

    gmodel = GestureConfig(
        path.join(model_folder, "model.pickle"), 
        path.join(model_folder, "encoder.pickle"),
        "test",
        datetime.datetime.now()
    )

    tskin = TSkin(TSkinConfig("C0:83:4F:34:28:38", Hand.RIGHT, "TEST", None, gmodel), True)
    tskin.start()

    print("connecting tskin", tskin)

    while not tskin.connected:
        pass

    i = 0

    while True:
        if not tskin.connected:
            print("Connecting...")
            time.sleep(0.2)
            continue

        if i > 5:
            break

        t = tskin.touch
        g = tskin.gesture

        if g:
            print(g)

        # print(tskin.sleep, tskin.gesture, tskin.acceleration, tskin.gyro, tskin.angle)

        if t and t.one_finger == OneFingerGesture.TAP_AND_HOLD:
            i += 1

        time.sleep(tskin.TICK)

    print("exit")

    tskin.terminate()


if __name__ == "__main__":
    test()


import multiprocessing
import json
import tactigon_gear

from os import path

from .client import data_collection as dc
from .client.client import Client
from .models.client import ClientConfig, UserData, DataCollectionConfig, HAL


def main():
    """
    Starting point of the main program
    :return: None
    """

    SDK_version = tactigon_gear.__version__

    while True:
        print("T-Gear SDK Client (v: {}), select an option:".format(SDK_version))
        print("Enter 0 to test server connection")
        print("Enter 1 for data collection")
        print("Enter 2 to send data to server")
        print("Enter 3 to train a new model")
        print("Enter 4 to download model from server")
        print("Type anything to EXIT")

        x = input()

        if x == "0":
            print("SERVER CONNECTION TEST")

            with open(path.join(path.dirname(__file__), "config_files", "user_data.json")) as user_data_file:
                user_data = UserData.FromJSON(json.load(user_data_file))
            
            with open(path.join(path.dirname(__file__), "config_files", "client.json")) as config_file:
                client_config = ClientConfig.FromJSON(json.load(config_file))

            with open(path.join(path.dirname(__file__), "config_files", "data_collection.json")) as dc_file:
                dc_config = DataCollectionConfig.FromJSON(json.load(dc_file))            
            
            client = Client(user_data, client_config, dc_config)
            client.test_connection()

        elif x == "1":
            print("DATA COLLECTION")

            with open(path.join(path.dirname(__file__), "config_files", "user_data.json")) as user_data_file:
                user_data = UserData.FromJSON(json.load(user_data_file))

            with open(path.join(path.dirname(__file__), "config_files", "hal.json")) as hal_file:
                hal_config = HAL.FromJSON(json.load(hal_file))

            with open(path.join(path.dirname(__file__), "config_files", "data_collection.json")) as dc_file:
                dc_config = DataCollectionConfig.FromJSON(json.load(dc_file))

            dc.collect(user_data, hal_config, dc_config)

        elif x == "2":
            print("SERVER DATA UPLOAD")
            with open(path.join(path.dirname(__file__), "config_files", "user_data.json")) as user_data_file:
                user_data = UserData.FromJSON(json.load(user_data_file))
            
            with open(path.join(path.dirname(__file__), "config_files", "client.json")) as config_file:
                client_config = ClientConfig.FromJSON(json.load(config_file))

            with open(path.join(path.dirname(__file__), "config_files", "data_collection.json")) as dc_file:
                dc_config = DataCollectionConfig.FromJSON(json.load(dc_file))

            client = Client(user_data, client_config, dc_config)
            client.send_data()
                        
        elif x == "3":
            print("SERVER MODEL PROCESSING")

            with open(path.join(path.dirname(__file__), "config_files", "user_data.json")) as user_data_file:
                user_data = UserData.FromJSON(json.load(user_data_file))
            
            with open(path.join(path.dirname(__file__), "config_files", "client.json")) as config_file:
                client_config = ClientConfig.FromJSON(json.load(config_file))

            with open(path.join(path.dirname(__file__), "config_files", "data_collection.json")) as dc_file:
                dc_config = DataCollectionConfig.FromJSON(json.load(dc_file))

            client = Client(user_data, client_config, dc_config)
            client.train_model()

        elif x == "4":
            print("SERVER MODEL DOWNLOAD")

            with open(path.join(path.dirname(__file__), "config_files", "user_data.json")) as user_data_file:
                user_data = UserData.FromJSON(json.load(user_data_file))
            with open(path.join(path.dirname(__file__), "config_files", "client.json")) as config_file:
                client_config = ClientConfig.FromJSON(json.load(config_file))

            with open(path.join(path.dirname(__file__), "config_files", "data_collection.json")) as dc_file:
                dc_config = DataCollectionConfig.FromJSON(json.load(dc_file))

            client = Client(user_data, client_config, dc_config)
            client.download_model()

        # elif x == "5":
        #     print("APP: SIMPLE GESTURE RECOGNITION")
        #     Application_Example.run()

        # elif x == "6":
        #     print("APP: GESTURE TO UI")
        #     Application_Gesture_to_UI.run()

        # elif x == "7":
        #     print("APP: GESTURE TO MQTT")
        #     Application_Gesture_to_MQTT.run()

        # elif x == "8":
        #     print("APP: FREECAD DEMO")
        #     Application_FreeCAD.run()
        
        # elif x == "9":
        #     print("APP: GESTURE TO IRONBOY DEMO")
        #     Application_Gesture_to_IronBoy.run()

        # elif x == "10":
        #     print("APP: GESTURE TO IRONBOY DEMO")
        #     Application_Gesture_to_IronBoy_ble.run()
            
        # elif x == "11":
        #     print("APP: ORCHESTRATOR")
        #     Application_Orchestrator.run()

        # elif x == "12":
        #     print("APP: RHINOCEROS")
        #     Application_Rhino.run()

        else:
            # exit from the loop
            break

    print("Process finished")


import setup_path
import time
import logging

from tactigon_gear.tskin_socket import TSkinSocket
from tactigon_gear.models.tskin import TSkinConfig, Hand, OneFingerGesture, TwoFingerGesture
from tactigon_gear.models.socket import SocketConfig
from tactigon_gear.models.audio import TSpeechObject, TSpeech, HotWord

logging.basicConfig(
    # filename="log.log", 
    level=logging.DEBUG
)
logging.getLogger("bleak").setLevel(logging.WARNING)

def get_tspeechobj() -> TSpeechObject:
    return TSpeechObject(
        [
            TSpeech(
                [HotWord("pick"), HotWord("place")],
                TSpeechObject(
                    [
                        TSpeech(
                            [HotWord("position")],
                            TSpeechObject(
                                [
                                    TSpeech(
                                        [HotWord("star"), HotWord("square")]
                                    )
                                ]
                            )
                        )                     
                    ]
                )
            )
        ]
    )

def main():
    cfg = TSkinConfig(
        "C0:83:3F:39:21:57",
        Hand.RIGHT
    )

    socket_cfg = SocketConfig("localhost", 50006)

    print("Init...")

    with TSkinSocket(cfg, socket_cfg) as tskin:
        print(tskin)
        while not tskin.connected:
            time.sleep(1)

        print("connected!")

        while True:
            if not tskin.connected:
                print("Reconnecting..")
                time.sleep(1)

            t = tskin.touch
            if t:
                if t.one_finger == OneFingerGesture.SINGLE_TAP:
                    print("record", tskin.record("testing.wav", 10))
                    # time.sleep(5)
                    # tskin.stop()
                elif t.two_finger == TwoFingerGesture.TWO_FINGER_TAP:
                    print("Listen", tskin._command, tskin.listen(get_tspeechobj()))
                    time.sleep(5)
                    tskin.stop()
                elif t.one_finger == OneFingerGesture.TAP_AND_HOLD:
                    break


            time.sleep(tskin.TICK)

if __name__ == "__main__":
    main()


__version__ = "5.5.0"

__all__ = [
    "TSkin", "TSkinSocket",
    "BLESelector",
    "TSkinConfig", "Hand", "Acceleration", "Angle", "Gyro",
    "TouchConfig", "OneFingerGesture", "TwoFingerGesture",
    "GestureConfig", "Gesture",
    "VoiceConfig", "TSpeechObject", "TSpeech", "HotWord", "Transcription",
    "SocketConfig"
]

from .tskin import TSkin
from .tskin_socket import TSkinSocket

from .models.ble import BLESelector
from .models.tskin import TSkinConfig, Hand, Acceleration, Angle, Gyro
from .models.touch import TouchConfig, OneFingerGesture, TwoFingerGesture
from .models.gesture import GestureConfig, Gesture
from .models.audio import VoiceConfig, TSpeechObject, TSpeech, HotWord, Transcription
from .models.socket import SocketConfig


from datetime import datetime
from dataclasses import dataclass

@dataclass
class Gesture:
    gesture: str
    probability: float
    confidence: float
    displacement: float

    def toJSON(self) -> dict:
        return {
            "gesture": self.gesture,
            "probability": self.probability,
            "confidence": self.confidence,
            "displacement": self.displacement,
        }

@dataclass
class GestureConfig:
    model_path: str
    encoder_path: str
    name: str
    created_at: datetime
    num_sample: int = 10
    gesture_prob_th: float = 0.85
    confidence_th: float = 5

    @classmethod
    def FromJSON(cls, json: dict):
        return cls(
            json["model_path"],
            json["encoder_path"],
            json["name"],
            datetime.fromisoformat(json["created_at"]),
            json["num_sample"] if "num_sample" in json and json["num_sample"] is not None else cls.num_sample,
            json["gesture_prob_th"] if "gesture_prob_th" in json and json["gesture_prob_th"] is not None else cls.gesture_prob_th,
            json["confidence_th"] if "confidence_th" in json and json["confidence_th"] is not None else cls.confidence_th,
            )
    
    def toJSON(self) -> dict:
        return {
            "model_path": self.model_path,
            "encoder_path": self.encoder_path,
            "name": self.name,
            "created_at": self.created_at.isoformat()
        }